from functools import wraps

from io import StringIO
import logging
import os
import re
import time

# email package declares these properties in the __all__ definition but
# pylint ignores that
from email import encoders, message_from_bytes  # pylint: disable=E0611
from email.charset import add_charset, QP  # pylint: disable=E0611
from email.generator import Generator  # pylint: disable=E0611
from email.header import Header, decode_header, make_header  # pylint: disable=E0611
from email.mime.base import MIMEBase  # pylint: disable=E0611
from email.mime.image import MIMEImage  # pylint: disable=E0611
from email.mime.multipart import MIMEMultipart  # pylint: disable=E0611
from email.mime.text import MIMEText  # pylint: disable=E0611

from pathlib import Path
from imaplib import IMAP4_SSL
from smtplib import SMTP, SMTP_SSL, ssl
from smtplib import SMTPConnectError, SMTPNotSupportedError, SMTPServerDisconnected

from typing import Any

from robot.libraries.BuiltIn import BuiltIn, RobotNotRunningError
from RPA.RobotLogListener import RobotLogListener


IMAGE_FORMATS = ["jpg", "jpeg", "bmp", "png", "gif"]
FLAG_DELETED = "\\Deleted"
FLAG_SEEN = "\\Seen"
FLAG_FLAGGED = "\\Flagged"

try:
    BuiltIn().import_library("RPA.RobotLogListener")
except RobotNotRunningError:
    pass


def imap_connection(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        if args[0].imap_conn is None:
            raise ValueError("Requires authorized IMAP connection")
        return f(*args, **kwargs)

    return wrapper


def smtp_connection(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        if args[0].smtp_conn is None:
            raise ValueError("Requires authorized SMTP connection")
        return f(*args, **kwargs)

    return wrapper


class ImapSmtp:
    """RPA Framework library providing SMTP and IMAP operations for
    specified mail server.

    Proper initialization needs `smtp_server` and `imap_server` defined.
    """

    def __init__(
        self,
        smtp_server: str = None,
        smtp_port: int = 587,
        imap_server: str = None,
        imap_port: int = 993,
        account: str = None,
        password: str = None,
    ) -> None:
        listener = RobotLogListener()
        listener.register_protected_keywords(
            ["RPA.Email.ImapSmtp.authorize", "RPA.Email.ImapSmtp.set_credentials"]
        )

        self.logger = logging.getLogger(__name__)
        self.smtp_server = smtp_server
        self.imap_server = imap_server if imap_server else smtp_server
        self.smtp_port = int(smtp_port)
        self.imap_port = int(imap_port)
        self.set_credentials(account, password)
        self.smtp_conn = None
        self.imap_conn = None

    def __del__(self) -> None:
        if self.smtp_conn:
            try:
                self.smtp_conn.quit()
            except SMTPServerDisconnected:
                self.logger.debug("Was already disconnected")
            finally:
                self.smtp_conn = None

        if self.imap_conn:
            self.select_folder()
            self.imap_conn.close()
            self.imap_conn.logout()
            self.imap_conn = None

    def set_credentials(self, account: str = None, password: str = None) -> None:
        """Set credentials

        :param account: user account as string, defaults to None
        :param password: user password as string, defaults to None

        Example:

        .. code-block:: robotframework

            Set Credentials   ${username}   ${password}
            Authorize

        """
        if account:
            self.account = account
        if password:
            self.password = password

    def authorize_smtp(
        self,
        account: str = None,
        password: str = None,
        smtp_server: str = None,
        smtp_port: int = None,
    ) -> None:
        """Authorize to SMTP server.

        Can be called without giving any parameters if library
        has been initialized with necessary information and/or
        keyword ``set_credentials`` has been used.

        :param account: SMTP account name, defaults to None
        :param password: SMTP account password, defaults to None
        :param smtp_server: SMTP server address, defaults to None
        :param smtp_port: SMTP server port, defaults to None (587 for SMTP)

        Example:

        .. code-block:: robotframework

            Authorize SMTP    ${username}   ${password}  smtp.gmail.com  587

        """
        if account is None and password is None:
            account = self.account
            password = self.password
        if smtp_server is None:
            smtp_server = self.smtp_server
        if smtp_port is None:
            smtp_port = self.smtp_port
        else:
            smtp_port = int(smtp_port)
        if smtp_server and account and password:
            try:
                self.smtp_conn = SMTP(smtp_server, smtp_port)
                self.send_smtp_hello()
                try:
                    self.smtp_conn.starttls()
                except SMTPNotSupportedError:
                    self.logger.warning("TLS not supported by the server")
            except SMTPConnectError:
                context = ssl.create_default_context()
                self.smtp_conn = SMTP_SSL(smtp_server, smtp_port, context=context)
            self.smtp_conn.login(account, password)
        else:
            self.logger.warning(
                "Server address, account and password are needed for "
                "authentication with SMTP"
            )
        if self.smtp_conn is None:
            self.logger.warning("Not able to establish SMTP connection")

    def authorize_imap(
        self,
        account: str = None,
        password: str = None,
        imap_server: str = None,
        imap_port: int = None,
    ) -> None:
        """Authorize to IMAP server.

        Can be called without giving any parameters if library
        has been initialized with necessary information and/or
        keyword ``set_credentials`` has been used.

        :param account: IMAP account name, defaults to None
        :param password: IMAP account password, defaults to None
        :param imap_server: IMAP server address, defaults to None
        :param imap_port: IMAP server port, defaults to None

        Example:

        .. code-block:: robotframework

            Authorize IMAP    ${username}   ${password}  imap.gmail.com  993

        """
        if account is None and password is None:
            account = self.account
            password = self.password
        if imap_server is None:
            imap_server = self.imap_server
        if imap_port is None:
            imap_port = self.imap_port
        else:
            imap_port = int(imap_port)
        if imap_server and account and password:
            self.imap_conn = IMAP4_SSL(imap_server, imap_port)
            self.imap_conn.login(account, password)
            self.imap_conn.select("INBOX")
        else:
            self.logger.warning(
                "Server address, account and password are needed for "
                "authentication with IMAP"
            )
        if self.imap_conn is None:
            self.logger.warning("Not able to establish IMAP connection")

    def authorize(
        self,
        account: str = None,
        password: str = None,
        smtp_server: str = None,
        imap_server: str = None,
        smtp_port: int = None,
        imap_port: int = None,
    ) -> None:
        # pylint: disable=C0301
        """Authorize user to SMTP and IMAP servers.

        Will use separately set credentials or those given in keyword call.

        :param account: user account as string, defaults to None
        :param password: user password as string, defaults to None
        :param smtp_server: SMTP server address, defaults to None
        :param imap_server: IMAP server address, defaults to None
        :param smtp_port: SMTP server port, defaults to None (587 for SMTP)
        :param imap_port: IMAP server port, defaults to None

        Example:

        .. code-block:: robotframework

            Authorize    ${username}   ${password}  smtp_server=smtp.gmail.com  smtp_port=587

        """  # noqa: E501
        self.authorize_smtp(account, password, smtp_server, smtp_port)
        self.authorize_imap(account, password, imap_server, imap_port)

    @smtp_connection
    def send_smtp_hello(self) -> None:
        """Send hello message to SMTP server.

        Required step when creating SMTP connection.

        .. code-block:: robotframework

            Send SMTP Hello

        """
        if self.smtp_conn:
            self.smtp_conn.ehlo()

    @smtp_connection
    def send_message(
        self,
        sender: str,
        recipients: str,
        subject: str = "",
        body: str = "",
        attachments: str = None,
        html: bool = False,
        images: str = None,
    ) -> bool:
        """Send SMTP email

        Valid sender values:
            - First Lastname <address@domain>
            - address@domain

        :param sender: who is sending, ie. 'from'
        :param recipients: who is receiving, ie. 'to'
        :param subject: mail subject field
        :param body: mail body content
        :param attachments: list of filepaths to attach, defaults to []
        :param html: if message content is in HTML, default `False`
        :param images: list of filepaths for inline use, defaults to []

        Example:

        .. code-block:: robotframework

            Send Message  sender@domain.com  recipient@domain.com
            ...           subject=Greetings Software Robot Developer
            ...           body=${email_body}
            ...           attachments:${CURDIR}${/}report.pdf

        """
        add_charset("utf-8", QP, QP, "utf-8")
        recipients, attachments, images = self._handle_message_parameters(
            recipients, attachments, images
        )
        msg = MIMEMultipart()

        self._add_attachments_to_msg(attachments, msg)

        msg["From"] = sender
        msg["To"] = ",".join(recipients)
        msg["Subject"] = Header(subject, "utf-8")

        if html:
            for im in images:
                im = im.strip()
                imname = Path(im).name
                body = body.replace(str(imname), f"cid:{imname}")
                with open(im, "rb") as f:
                    img = MIMEImage(f.read())
                    img.add_header("Content-ID", f"<{imname}>")
                    msg.attach(img)
            htmlpart = MIMEText(body, "html", "UTF-8")
            msg.attach(htmlpart)
        else:
            textpart = MIMEText(body, "plain", "UTF-8")
            msg.attach(textpart)
            for im in images:
                im = im.strip()
                imname = Path(im).name
                with open(im, "rb") as f:
                    img = MIMEImage(f.read())
                    msg.add_header(
                        "Content-Disposition",
                        f"inline; filename= {imname}",
                    )
                    msg.attach(img)

        # Create a generator and flatten message object to 'file’
        str_io = StringIO()
        g = Generator(str_io, False)
        g.flatten(msg)

        try:
            self.smtp_conn.sendmail(sender, recipients, str_io.getvalue())
        except Exception as err:
            raise ValueError(f"Send Message failed: {err}") from err
        return True

    def _handle_message_parameters(self, recipients, attachments, images):
        if attachments is None:
            attachments = []
        if images is None:
            images = []
        if not isinstance(recipients, list):
            recipients = recipients.split(",")
        if not isinstance(attachments, list):
            attachments = str(attachments).split(",")
        if not isinstance(images, list):
            images = str(images).split(",")
        return recipients, attachments, images

    def _add_attachments_to_msg(self, attachments: list = None, msg=None):
        if len(attachments) > 0:
            for filename in attachments:
                if os.path.dirname(filename) == "":
                    filename = str(Path.cwd() / filename)
                self.logger.debug("Adding attachment: %s", filename)
                with open(filename, "rb") as attachment:
                    _, ext = filename.lower().rsplit(".", 1)
                    if ext in IMAGE_FORMATS:
                        # image attachment
                        part = MIMEImage(
                            attachment.read(),
                            name=Path(filename).name,
                            _subtype=ext,
                        )
                    else:
                        # attach other filetypes
                        part = MIMEBase("application", "octet-stream")
                        part.set_payload(attachment.read())
                        encoders.encode_base64(part)
                    part.add_header(
                        "Content-Disposition",
                        f"attachment; filename= {Path(filename).name}",
                    )
                    msg.attach(part)

    @imap_connection
    def _fetch_messages(self, mail_ids: list) -> list:
        messages = []
        for mail_id in mail_ids:
            _, data = self.imap_conn.fetch(mail_id, "(RFC822)")
            message = message_from_bytes(data[0][1])
            message_dict = {"Mail-Id": mail_id, "Message": message}
            for k, v in message.items():
                msg_item = decode_header(v)
                message_dict[k] = make_header(msg_item)
            message_dict["Body"], has_attachments = self._get_decoded_email_body(
                message
            )
            message_dict["Has-Attachments"] = has_attachments
            messages.append(message_dict)
        return messages

    def _get_decoded_email_body(self, message):
        """Decode email body.

        Detect character set if the header is not set.
        We try to get text/plain, but if there is not one then fallback to text/html.

        :param message_body: Raw 7-bit message body input e.g. from imaplib. Double
         encoded in quoted-printable and latin-1
        :return: Message body as unicode string and information if message has
         attachments
        """
        text = ""
        has_attachments = False
        if message.is_multipart():
            html = None

            for part in message.walk():
                # content_maintype = part.get_content_maintype()
                content_disposition = part.get("Content-Disposition")
                if content_disposition and "attachment" in content_disposition:
                    has_attachments = True
                    continue
                if part.get_content_charset() is None:
                    # We cannot know the character set, so return decoded "something"
                    text = part.get_payload(decode=True)
                    continue

                charset = part.get_content_charset()

                if part.get_content_type() == "text/plain":
                    text = str(
                        part.get_payload(decode=True), str(charset), "ignore"
                    ).encode("utf8", "replace")

                if part.get_content_type() == "text/html":
                    html = str(
                        part.get_payload(decode=True), str(charset), "ignore"
                    ).encode("utf8", "replace")

            if text:
                return (
                    (text.strip(), has_attachments) if text else ("", has_attachments)
                )
            else:
                return (
                    (html.strip(), has_attachments) if html else ("", has_attachments)
                )
        else:
            text = str(
                message.get_payload(decode=True),
                message.get_content_charset(),
                "ignore",
            ).encode("utf8", "replace")
            return text.strip(), has_attachments

    @imap_connection
    def _search_message(self, criterion: str) -> list:
        return self.imap_conn.search(None, "(" + criterion + ")")

    def _search_and_return_mail_ids(self, criterion: str) -> list:
        _, data = self._search_message(criterion)
        mail_ids = bytes.decode(data[0])
        return mail_ids.split() if len(mail_ids) > 1 else []

    @imap_connection
    def _set_message_flag(
        self, mail_id: str = None, flag: str = None, unset: bool = False
    ):
        flag_type = "-FLAGS" if unset else "+FLAGS"
        # self.imap_conn.uid("STORE", mail_id, flag_type, f"({flag})")
        mail_uid = self._fetch_uid(mail_id)
        status_code, data = self.imap_conn.uid("STORE", mail_uid, flag_type, flag)
        data_message = data[0]
        self.logger.debug("Set message flag: %s, %s", status_code, data)
        if status_code != "OK":
            self.logger.debug(
                "Message flag '%s' failed for mail_id: %s, mail_uid: %s",
                flag,
                mail_id,
                mail_uid,
            )
        return status_code == "OK" and data_message is not None

    @imap_connection
    def _delete_messages(self, mail_ids: list) -> None:
        for mail_id in mail_ids:
            self._set_message_flag(mail_id, FLAG_DELETED)
        self.imap_conn.expunge()

    @imap_connection
    def delete_message(self, criterion: str = "") -> bool:
        """Delete single message from server based on criterion.

        If criterion does not return exactly 1 message then delete is not done.

        :param criterion: filter messages based on this, defaults to ""
        :return: True if success, False if not

        Example:

        .. code-block:: robotframework

            Delete Message  SUBJECT \"Greetings RPA developer\"

        """
        self._validate_criterion(criterion)
        _, data = self._search_message(criterion)
        mail_ids = data[0].split()
        if len(mail_ids) != 1:
            self.logger.warning(
                "Delete message criteria matched %d messages. Not deleting.",
                len(mail_ids),
            )
            return False
        else:
            self._delete_messages(mail_ids)
            return True

    @imap_connection
    def delete_messages(self, criterion: str = "") -> bool:
        """Delete messages from server based on criterion.

        :param criterion: filter messages based on this, defaults to ""
        :return: True if success, False if not

        Example:

        .. code-block:: robotframework

            Delete Messages  SUBJECT Greetings

        """
        self._validate_criterion(criterion)
        mail_ids = self._search_and_return_mail_ids(criterion)
        self._delete_messages(mail_ids)
        return True

    @imap_connection
    def save_messages(self, criterion: str = "", target_folder: str = None) -> bool:
        # pylint: disable=C0301
        """Save messages based on criteria and store them to target folder
        with attachment files.

        Does not save message if `target_folder` is not given.

        :param criterion: filter messages based on this, defaults to ""
        :param target_folder: path to folder where message are saved, defaults to None
        :return: True if success, False if not

        Example:

        .. code-block:: robotframework

            Save Messages  SUBJECT Important message  target_folder=${USERDIR}${/}messages

        """  # noqa: E501
        self._validate_criterion(criterion)
        if target_folder is None:
            target_folder = os.path.expanduser("~")
        _, data = self._search_message(criterion)
        mail_ids = data[0].split()
        for mail_id in mail_ids:
            _, data = self.imap_conn.fetch(mail_id, "(RFC822)")
            emlfile = Path(target_folder) / f"{str(int(mail_id))}.eml"
            with open(emlfile, "wb") as f:
                f.write(data[0][1])
        return True

    @imap_connection
    def list_messages(self, criterion: str = "") -> Any:
        """Return list of messages matching criterion.

        :param criterion: list emails matching this, defaults to ""
        :return: list of messages or False

        Example:

        .. code-block:: robotframework

            @{emails}  List Messages  SUBJECT \"rpa task\"
            FOR  ${email}  IN  @{EMAILS}
                Log  ${email}[Subject]
                Log  ${email}[From]
                Log  ${email}[Date]
                Log  ${email}[Delivered-To]
                Log  ${email}[Received]
                Log  ${email}[Has-Attachments]
            END

        """
        self.logger.info("List messages: %s", criterion)
        _, data = self._search_message(criterion)
        mail_ids = data[0].split()
        return self._fetch_messages(mail_ids)

    @imap_connection
    def save_attachments(
        self, criterion: str = "", target_folder: str = None, overwrite: bool = False
    ) -> Any:
        # pylint: disable=C0301
        """Save mail attachments into local folder.

        :param criterion: attachments are saved for mails matching this, defaults to ""
        :param target_folder: local folder for saving attachments to (needs to exist),
            defaults to user's home directory if None
        :param overwrite: overwrite existing file is True, defaults to False
        :return: list of saved attachments or False

        Example:

        .. code-block:: robotframework

            ${numsaved}  Save Attachments   SUBJECT \"rpa task\"
            ...          target_folder=${CURDIR}${/}messages  overwrite=True

        """  # noqa: E501
        attachments_saved = []
        messages = self.list_messages(criterion)
        for msg in messages:
            attachments_saved.append(
                self.save_attachment(msg, target_folder, overwrite)
            )
        return attachments_saved if len(attachments_saved) > 0 else False

    def save_attachment(self, message, target_folder, overwrite):
        # pylint: disable=C0301
        """Save mail attachment into local folder

        :param message: message item
        :param target_folder: local folder for saving attachments to (needs to exist),
            defaults to user's home directory if None
        :param overwrite: overwrite existing file is True, defaults to False

        Example:

        .. code-block:: robotframework

            @{emails}  List Messages  SUBJECT \"rpa task\"
            FOR  ${email}  IN  @{EMAILS}
                Run Keyword If   ${email}[Has-Attachments] == True
                ...              Save Attachment  ${email}  target_folder=${CURDIR}  overwrite = True
            END

        """  # noqa: E501
        if target_folder is None:
            target_folder = os.path.expanduser("~")
        msg = message["Message"] if isinstance(message, dict) else message
        attachments_saved = []
        for part in msg.walk():
            content_maintype = part.get_content_maintype()
            content_disposition = part.get("Content-Disposition")
            if content_maintype != "multipart" and content_disposition is not None:
                filename = part.get_filename()
                self.logger.info("%s %s", filename, content_maintype)
                if bool(filename):
                    filepath = Path(target_folder) / filename
                    if not filepath.exists() or overwrite:
                        with open(filepath, "wb") as f:
                            f.write(part.get_payload(decode=True))
                            attachments_saved.append(filepath)

    @imap_connection
    def wait_for_message(
        self, criterion: str = "", timeout: float = 5.0, interval: float = 1.0
    ) -> Any:
        """Wait for email matching `criterion` to arrive into mailbox.

        :param criterion: message filter to wait for, defaults to ""
        :param timeout: total time in seconds to wait for email, defaults to 5.0
        :param interval: time in seconds for new check, defaults to 1.0
        :return: list of messages or False

        Example:

        .. code-block:: robotframework

            @{emails}  Wait For Message  SUBJECT \"rpa task\"  timeout=300  interval=10

        """
        self._validate_criterion(criterion)
        end_time = time.time() + float(timeout)
        while time.time() < end_time:
            self.imap_conn.select("inbox")
            _, data = self._search_message(criterion)
            mail_ids = data[0].split()
            self.logger.warning(
                "Wait for message found matching message %d time(s).", len(mail_ids)
            )
            if len(mail_ids) > 0:
                return self._fetch_messages(mail_ids)
            time.sleep(interval)
        return False

    def _parse_folders(self, folders):
        parsed_folders = []
        folder_regex = r'\((?P<flags>.*)\)."(?P<delimiter>.*)"."(?P<name>.*)".*'
        for f in folders:
            flags, delimiter, name = re.search(folder_regex, bytes.decode(f)).groups()

            parsed_folders.append(
                {"name": name, "flags": flags, "delimiter": delimiter}
            )
        return parsed_folders

    @imap_connection
    def get_folder_list(self, subdirectory: str = None, pattern: str = None) -> list:
        """Get list of folders on the server

        :param subdirectory: list subdirectories for this folder
        :param pattern: list folders matching this pattern
        :return: list of folders

        Example:

        .. code-block:: robotframework

            @{folders}  Get Folder List
            @{folders}  Get Folder List  pattern=important
            @{folders}  Get Folder List  subdirectory=sub

        """
        self.logger.info("Get folder list")
        kwparams = {}
        if subdirectory:
            kwparams["directory"] = subdirectory
        if pattern:
            kwparams["pattern"] = pattern

        status_code, folders = self.imap_conn.list(**kwparams)
        if status_code == "OK":
            return self._parse_folders(folders)
        else:
            return []

    @imap_connection
    def select_folder(self, folder_name: str = "INBOX") -> int:
        """Select folder by name

        Returns number of messages in the folder or
        exception if folder does not exist on the server.

        :param folder_name: name of the folder to select
        :return: message count in the selected folder

        Example:

        .. code-block:: robotframework

            Select Folder   subfolder

        """
        self.logger.info("Select folder: %s", folder_name)
        status_code, data = self.imap_conn.select(mailbox=folder_name, readonly=False)
        if status_code == "OK":
            message_count = bytes.decode(data[0])
            return int(message_count)
        else:
            raise ValueError("Folder '%s' does not exist on the server" % folder_name)

    @imap_connection
    def rename_folder(
        self, oldname: str = None, newname: str = None, suppress_error: bool = False
    ) -> bool:
        """Rename email folder

        :param oldname: current folder name
        :param newname: new name for the folder
        :param suppress_error: to silence warning message, defaults to False
        :return: True if operation was successful, False if not

        Example:

        .. code-block:: robotframework

            Rename Folder   subfolder   filtered

        """
        if oldname is None or newname is None:
            raise KeyError(
                "Both 'oldname' and 'newname' and required for rename folder"
            )
        self.logger.info("Rename folder '%s' to '%s'", oldname, newname)
        status_code, data = self.imap_conn.rename(oldname, newname)

        if status_code == "OK":
            return True
        else:
            if suppress_error is False:
                self.logger.warning(
                    "Folder rename failed with message: '%s'",
                    bytes.decode(data[0]),
                )
            return False

    @imap_connection
    def delete_folder(self, folder_name: str = None) -> bool:
        """Delete email folder

        :param folder_name: current folder name
        :return: True if operation was successful, False if not

        Example:

        .. code-block:: robotframework

            Delete Folder   filtered

        """
        if folder_name is None:
            raise KeyError("'folder_name' is required for delete folder")
        self.logger.info("Delete folder '%s'", folder_name)
        status_code, data = self.imap_conn.delete(folder_name)
        if status_code == "OK":
            return True
        else:
            self.logger.warning(
                "Delete folder '%s' response status: '%s %s'",
                folder_name,
                status_code,
                bytes.decode(data[0]),
            )
            return False

    @imap_connection
    def create_folder(self, folder_name: str = None) -> bool:
        """Create email folder

        :param folder_name: name for the new folder
        :return: True if operation was successful, False if not

        Example:

        .. code-block:: robotframework

            Create Folder   filtered

        """
        if folder_name is None:
            raise KeyError("'folder_name' is required for create folder")
        self.logger.info("Create folder '%s'", folder_name)
        status_code, data = self.imap_conn.create(folder_name)
        if status_code == "OK":
            return True
        else:
            self.logger.warning(
                "Create folder '%s' response status: '%s %s'",
                folder_name,
                status_code,
                bytes.decode(data[0]),
            )
            return False

    @imap_connection
    def flag_messages(self, criterion: str = None, unflag: bool = False) -> Any:
        """Mark messages as `flagged`

        :param criterion: mark messages matching criterion
        :param unflag: to mark messages as not `flagged`
        :return: successful operations (int), matching messages (int)

        Example:

        .. code-block:: robotframework

            ${flagged}  ${oftotal}    Flag Messages   SUBJECT rpa
            ${unflagged}  ${oftotal}  Flag Messages   SUBJECT rpa  unflag=True

        """
        self._validate_criterion(criterion)
        if unflag:
            self.logger.info("Unflag messages: '%s'", criterion)
        else:
            self.logger.info("Flag messages: '%s'", criterion)
        mail_ids = self._search_and_return_mail_ids(criterion)
        success_count = 0
        for mail_id in mail_ids:
            status = self._set_message_flag(mail_id, FLAG_FLAGGED, unflag)
            success_count += 1 if status else 0
        return success_count, len(mail_ids)

    @imap_connection
    def unflag_messages(self, criterion: str = None) -> Any:
        """Mark messages as not `flagged`

        :param criterion: mark messages matching criterion
        :return: successful operations (int), matching messages (int)

        Example:

        .. code-block:: robotframework

            ${unflagged}  ${oftotal}  Unflag Messages   SUBJECT rpa

        """
        return self.flag_messages(criterion, unflag=True)

    @imap_connection
    def mark_as_read(self, criterion: str = None, unread: bool = False) -> Any:
        """Mark messages as `read`

        :param criterion: mark messages matching criterion
        :param unread: to mark messages as not `read`
        :return: successful operations (int), matching messages (int)

        Example:

        .. code-block:: robotframework

            ${read}  ${oftotal}  Mark As Read   SUBJECT rpa

        """
        self._validate_criterion(criterion)
        if unread:
            self.logger.info("Mark messages as unread: '%s'", criterion)
        else:
            self.logger.info("Mark messages as read: '%s'", criterion)
        mail_ids = self._search_and_return_mail_ids(criterion)
        success_count = 0
        for mail_id in mail_ids:
            status = self._set_message_flag(mail_id, FLAG_SEEN, unread)
            success_count += 1 if status else 0
        return success_count, len(mail_ids)

    @imap_connection
    def mark_as_unread(self, criterion: str = None) -> Any:
        """Mark messages as not `read`

        :param criterion: mark messages matching criterion
        :return: successful operations (int), matching messages (int)

        Example:

        .. code-block:: robotframework

            ${unread}  ${oftotal}  Mark As Unread   SUBJECT rpa

        """
        return self.mark_as_read(criterion, unread=True)

    def _validate_criterion(self, criterion: str) -> bool:
        if criterion is None or len(criterion) < 1:
            raise KeyError("Criterion is required parameter")

    def _fetch_uid(self, mail_id):
        _, data = self.imap_conn.fetch(mail_id, "(UID)")
        pattern_uid = re.compile(r"\d+ \(UID (?P<uid>\d+)\)")
        match = pattern_uid.match(bytes.decode(data[0]))
        return match.group("uid")



import logging
from pathlib import Path
import time
from typing import Any

from exchangelib import (
    Account,
    Configuration,
    Credentials,
    DELEGATE,
    EWSDateTime,
    EWSTimeZone,
    FileAttachment,
    Folder,
    HTMLBody,
    IMPERSONATION,
    Mailbox,
    Message,
)


class Exchange:
    """Library for interfacing with Microsoft Exchange Web Services (EWS)."""

    def __init__(self) -> None:
        self.logger = logging.getLogger(__name__)
        self.credentials = None
        self.config = None
        self.account = None

    def authorize(
        self,
        username: str,
        password: str,
        autodiscover: bool = True,
        access_type: str = "DELEGATE",
        server: str = None,
        primary_smtp_address: str = None,
    ) -> None:
        """Connect to Exchange account

        :param username: account username
        :param password: account password
        :param autodiscover: use autodiscover or set it off
        :param accesstype: default "DELEGATE", other option "IMPERSONATION"
        :param server: required for configuration options
        :param primary_smtp_address: by default set to username, but can be
            set to be different than username
        """
        kwargs = {}
        kwargs["autodiscover"] = autodiscover
        kwargs["access_type"] = (
            DELEGATE if access_type.upper() == "DELEGATE" else IMPERSONATION
        )
        kwargs["primary_smtp_address"] = (
            primary_smtp_address if primary_smtp_address else username
        )
        self.credentials = Credentials(username, password)
        if server:
            self.config = Configuration(server=server, credentials=self.credentials)
            kwargs["config"] = self.config
        else:
            kwargs["credentials"] = self.credentials

        self.account = Account(**kwargs)

    def list_messages(self, folder_name: str = None, count: int = 100) -> list:
        """List messages in the account inbox. Order by descending
        received time.

        :param count: number of messages to list
        """
        # pylint: disable=no-member
        messages = []
        source_folder = self._get_folder_object(folder_name)
        for item in source_folder.all().order_by("-datetime_received")[:count]:
            messages.append(
                {
                    "subject": item.subject,
                    "sender": item.sender,
                    "datetime_received": item.datetime_received,
                    "body": item.body,
                }
            )
        return messages

    def _get_all_items_in_folder(self, folder_name=None, parent_folder=None) -> list:
        if parent_folder is None or parent_folder is self.account.inbox:
            target_folder = self.account.inbox / folder_name
        else:
            target_folder = self.account.inbox / parent_folder / folder_name
        return target_folder.all()

    def send_message(
        self,
        recipients: str,
        subject: str = "",
        body: str = "",
        attachments: str = None,
        html: bool = False,
        images: str = None,
        cc: str = None,
        bcc: str = None,
        save: bool = False,
    ):
        """Keyword for sending message through connected Exchange account.

        Email addresses can be prefixed with `ex:` to indicate Exchange
        account address.

        Recipients is ``required`` parameter.

        :param recipients: list of email addresses, defaults to []
        :param subject: message subject, defaults to ""
        :param body: message body, defaults to ""
        :param attachments: list of filepaths to attach, defaults to []
        :param html: if message content is in HTML, default `False`
        :param images: list of filepaths for inline use, defaults to []
        :param cc: list of email addresses, defaults to []
        :param bcc: list of email addresses, defaults to []
        :param save: is sent message saved to Sent messages folder or not,
            defaults to False
        """
        recipients, cc, bcc, attachments, images = self._handle_message_parameters(
            recipients, cc, bcc, attachments, images
        )
        self.logger.info("Sending message to %s", ",".join(recipients))

        m = Message(
            account=self.account,
            subject=subject,
            body=body,
            to_recipients=recipients,
            cc_recipients=cc,
            bcc_recipients=bcc,
        )

        self._add_attachments_to_msg(attachments, m)
        self._add_images_inline_to_msg(images, html, body, m)

        if html:
            m.body = HTMLBody(body)
        else:
            m.body = body

        if save:
            m.folder = self.account.sent
            m.send_and_save()
        else:
            m.send()
        return True

    def _handle_message_parameters(self, recipients, cc, bcc, attachments, images):
        if cc is None:
            cc = []
        if bcc is None:
            bcc = []
        if attachments is None:
            attachments = []
        if images is None:
            images = []
        if not isinstance(recipients, list):
            recipients = recipients.split(",")
        if not isinstance(cc, list):
            cc = [cc]
        if not isinstance(bcc, list):
            bcc = [bcc]
        if not isinstance(attachments, list):
            attachments = str(attachments).split(",")
        if not isinstance(images, list):
            images = str(images).split(",")
        recipients, cc, bcc = self._handle_recipients(recipients, cc, bcc)
        return recipients, cc, bcc, attachments, images

    def _handle_recipients(self, recipients, cc, bcc):
        recipients = [
            Mailbox(email_address=p.split("ex:")[1]) if "ex:" in p else p
            for p in recipients
        ]
        cc = [Mailbox(email_address=p.split("ex:")[1]) if "ex:" in p else p for p in cc]
        bcc = [
            Mailbox(email_address=p.split("ex:")[1]) if "ex:" in p else p for p in bcc
        ]
        return recipients, cc, bcc

    def _add_attachments_to_msg(self, attachments, msg):
        for attachment in attachments:
            attachment = attachment.strip()
            with open(attachment, "rb") as f:
                atname = str(Path(attachment).name)
                fileat = FileAttachment(name=atname, content=f.read())
                msg.attach(fileat)

    def _add_images_inline_to_msg(self, images, html, body, msg):
        for image in images:
            image = image.strip()
            with open(image, "rb") as f:
                imname = str(Path(image).name)
                fileat = FileAttachment(
                    name=imname, content=f.read(), content_id=imname
                )
                msg.attach(fileat)
                if html:
                    body = body.replace(imname, f"cid:{imname}")

    def create_folder(self, folder_name: str = None, parent_folder: str = None) -> bool:
        """Create email folder

        :param folder_name: name for the new folder
        :param parent_folder: name for the parent folder, by default INBOX
        :return: True if operation was successful, False if not
        """
        if folder_name is None:
            raise KeyError("'folder_name' is required for create folder")
        if parent_folder is None or parent_folder is self.account.inbox:
            parent = self.account.inbox
        else:
            parent = self.account.inbox / parent_folder / folder_name
        self.logger.info(
            "Create folder '%s'",
            folder_name,
        )
        new_folder = Folder(parent=parent, name=folder_name)
        new_folder.save()

    def delete_folder(self, folder_name: str = None, parent_folder: str = None) -> bool:
        """Delete email folder

        :param folder_name: current folder name
        :param parent_folder: name for the parent folder, by default INBOX
        :return: True if operation was successful, False if not
        """
        if folder_name is None:
            raise KeyError("'folder_name' is required for delete folder")
        if parent_folder is None or parent_folder is self.account.inbox:
            folder_to_delete = self.account.inbox / folder_name
        else:
            folder_to_delete = self.account.inbox / parent_folder / folder_name
        self.logger.info(
            "Delete folder  '%s'",
            folder_name,
        )
        folder_to_delete.delete()

    def rename_folder(
        self, oldname: str = None, newname: str = None, parent_folder: str = None
    ) -> bool:
        """Rename email folder

        :param oldname: current folder name
        :param newname: new name for the folder
        :param parent_folder: name for the parent folder, by default INBOX
        :return: True if operation was successful, False if not
        """
        if oldname is None or newname is None:
            raise KeyError("'oldname' and 'newname' are required for rename folder")
        if parent_folder is None or parent_folder is self.account.inbox:
            parent = self.account.inbox
        else:
            parent = self.account.inbox / parent_folder
        self.logger.info(
            "Rename folder '%s' to '%s'",
            oldname,
            newname,
        )
        items = self._get_all_items_in_folder(oldname, parent_folder)
        old_folder = Folder(parent=parent, name=oldname)
        old_folder.name = newname
        old_folder.save()
        items.move(to_folder=parent / newname)
        self.delete_folder(oldname, parent_folder)

    def empty_folder(
        self,
        folder_name: str = None,
        parent_folder: str = None,
        delete_sub_folders: bool = False,
    ) -> bool:
        """Empty email folder of all items

        :param folder_name: current folder name
        :param parent_folder: name for the parent folder, by default INBOX
        :param delete_sub_folders: delete sub folders or not, by default False
        :return: True if operation was successful, False if not
        """
        if folder_name is None:
            raise KeyError("'folder_name' is required for empty folder")
        if parent_folder is None or parent_folder is self.account.inbox:
            empty_folder = self.account.inbox / folder_name
        else:
            empty_folder = self.account.inbox / parent_folder / folder_name
        self.logger.info("Empty folder '%s'", folder_name)
        empty_folder.empty(delete_sub_folders=delete_sub_folders)

    def move_messages(
        self,
        criterion: str = "",
        source: str = None,
        target: str = None,
        contains: bool = False,
    ) -> bool:
        """Move message(s) from source folder to target folder

        Criterion examples:

            - subject:my message subject
            - body:something in body
            - sender:sender@domain.com

        :param criterion: move messages matching this criterion
        :param source: source folder
        :param target: target folder
        :param contains: if matching should be done using `contains` matching
             and not `equals` matching, default `False` is means `equals` matching
        :return: boolean result of operation, True if 1+ items were moved else False
        """
        source_folder = self._get_folder_object(source)
        target_folder = self._get_folder_object(target)
        if source_folder == target_folder:
            raise KeyError("Source folder is same as target folder")
        self.logger.info("Move messages")
        filter_dict = self._get_filter_key_value(criterion, contains)
        items = source_folder.filter(**filter_dict)
        if items and items.count() > 0:
            items.move(to_folder=target_folder)
            return True
        else:
            self.logger.warning("No items match criterion '%s'", criterion)
            return False

    def _get_folder_object(self, folder_name):
        if folder_name is None or folder_name == "INBOX":
            folder_object = self.account.inbox
        elif "INBOX" in folder_name:
            folder_object = self.account.inbox / folder_name.replace(
                "INBOX/", ""
            ).replace("INBOX /", "")
        else:
            folder_object = self.account.inbox / folder_name
        return folder_object

    def _get_filter_key_value(self, criterion, contains):
        if criterion.startswith("subject:"):
            search_key = "subject"
        elif criterion.startswith("body:"):
            search_key = "body"
        elif criterion.startswith("sender:"):
            search_key = "sender"
        else:
            raise KeyError("Unknown criterion for filtering items '%s'" % criterion)
        if contains:
            search_key += "__contains"
        _, search_val = criterion.split(":", 1)
        return {search_key: search_val}

    def wait_for_message(
        self,
        criterion: str = "",
        timeout: float = 5.0,
        interval: float = 1.0,
        contains: bool = False,
    ) -> Any:
        """Wait for email matching `criterion` to arrive into INBOX.

        :param criterion: wait for message matching criterion
        :param timeout: total time in seconds to wait for email, defaults to 5.0
        :param interval: time in seconds for new check, defaults to 1.0
        :param contains: if matching should be done using `contains` matching
             and not `equals` matching, default `False` is means `equals` matching
        :return: list of messages
        """
        self.logger.info("Wait for messages")
        end_time = time.time() + float(timeout)
        filter_dict = self._get_filter_key_value(criterion, contains)
        items = None
        tz = EWSTimeZone.localzone()
        right_now = tz.localize(EWSDateTime.now())  # pylint: disable=E1101
        while time.time() < end_time:
            items = self.account.inbox.filter(  # pylint: disable=E1101
                **filter_dict, datetime_received__gte=right_now
            )
            if items.count() > 0:
                break
            time.sleep(interval)
        messages = []
        for item in items:
            messages.append(
                {
                    "subject": item.subject,
                    "sender": item.sender,
                    "datetime_received": item.datetime_received,
                    "body": item.body,
                }
            )
        if len(messages) == 0:
            self.logger.info("Did not receive any matching items")
        return messages



import logging
import sys
import time
from dataclasses import dataclass, astuple
from pathlib import Path

from PIL import Image
from PIL import ImageDraw
from PIL import ImageOps

try:
    # Attempt to enable DPI awareness, for platforms that support it.
    # Needs to be enabled `before` pyscreenshot is imported.
    import ctypes

    ctypes.windll.user32.SetProcessDPIAware()
except AttributeError:
    pass

import pyscreenshot as ImageGrab
from RPA.core.notebook import notebook_image

try:
    # Check if opencv is available,
    # for improved template matching
    import cv2
    import numpy

    HAS_OPENCV = True
except ImportError:
    HAS_OPENCV = False


def clamp(minimum, value, maximum):
    """Clamp value between given minimum and maximum."""
    return max(minimum, min(value, maximum))


def chunks(obj, size, start=0):
    """Convert `obj` container to list of chunks of `size`."""
    return [obj[i : i + size] for i in range(start, len(obj), size)]


def to_image(obj):
    """Convert `obj` to instance of Pillow's Image class."""
    if obj is None or isinstance(obj, Image.Image):
        return obj
    return Image.open(obj)


def to_point(obj):
    """Convert `obj` to instance of Point."""
    if obj is None or isinstance(obj, Point):
        return obj
    if isinstance(obj, str):
        obj = obj.split(",")
    return Point(*(int(i) for i in obj))


def to_region(obj):
    """Convert `obj` to instance of Region."""
    if obj is None or isinstance(obj, Region):
        return obj
    if isinstance(obj, str):
        obj = obj.split(",")
    return Region(*(int(i) for i in obj))


@dataclass
class Point:
    """Container for a 2D point."""

    x: int
    y: int

    def as_tuple(self):
        return astuple(self)


@dataclass
class Region:
    """Container for a 2D rectangular region."""

    left: int
    top: int
    right: int
    bottom: int

    def __post_init__(self):
        if self.left >= self.right:
            raise ValueError("Invalid width")
        if self.top >= self.bottom:
            raise ValueError("Invalid height")

    @classmethod
    def from_size(cls, x, y, width, height):
        return cls(x, y, x + width, y + height)

    @property
    def width(self):
        return self.right - self.left

    @property
    def height(self):
        return self.bottom - self.top

    @property
    def area(self):
        return self.width * self.height

    @property
    def center(self):
        return Point(
            x=int((self.left + self.right) / 2), y=int((self.top + self.bottom) / 2)
        )

    def as_tuple(self):
        return astuple(self)

    def move(self, left, top):
        width, height = self.width, self.height
        self.left = clamp(0, self.left + left, sys.maxsize)
        self.top = clamp(0, self.top + top, sys.maxsize)
        self.right = self.left + width
        self.bottom = self.top + height


@dataclass
class RGB:
    """Container for a single RGB value."""

    red: int
    green: int
    blue: int

    @classmethod
    def from_pixel(cls, value):
        """Create RGB value from pillow getpixel() return value."""
        # RGB(A), ignore possible alpha channel
        if isinstance(value, tuple):
            red, green, blue = value[:3]
        # Grayscale
        else:
            red, green, blue = [value] * 3

        return cls(red, green, blue)

    def luminance(self):
        """Approximate (perceived) luminance for RGB value."""
        return (self.red * 2 + self.green * 3 + self.blue) // 6


class ImageNotFoundError(Exception):
    """Raised when template matching fails."""


class Images:
    """Library for taking screenshots, matching templates, and
    manipulating images.
    """

    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.matcher = TemplateMatcher(opencv=HAS_OPENCV)

    def take_screenshot(self, filename=None, region=None, save_format="PNG"):
        """Take a screenshot of the current desktop.

        :param filename:    Save screenshot to filename
        :param region:      Region to crop screenshot to
        :param save_format: File format to save the screenshot in
        """
        region = to_region(region)

        if region is not None:
            image = ImageGrab.grab(bbox=region.as_tuple())
        else:
            image = ImageGrab.grab()

        if filename is not None:
            filename = Path(filename).with_suffix(f".{save_format.lower()}")
            image.save(filename, save_format)
            notebook_image(filename)
            self.logger.info("Saved screenshot as '%s'", filename)

        return image

    def crop_image(self, image, region, filename=None, save_format="PNG"):
        """Take a screenshot of the current desktop.

        :param image:       Image to crop
        :param region:      Region to crop image to
        :param filename:    Save cropped image to filename
        :param save_format: File format to save the image in
        """
        region = to_region(region)
        image = to_image(image)

        image = image.crop(region.as_tuple())
        image.load()

        if filename:
            image.save(filename, save_format)
            notebook_image(filename)

    def find_template_in_image(
        self, image, template, region=None, limit=None, tolerance=None
    ):
        """Attempt to find the template from the given image.

        :param image:       Path to image or Image instance, used to search from
        :param template:    Path to image or Image instance, used to search with
        :param limit:       Limit returned results to maximum of `limit`.
        :param region:      Area to search from. Can speed up search significantly.
        :param tolerance:   Tolerance for matching, value between 0.1 and 1.0
        :return:            List of matching regions
        :raises ImageNotFoundError: No match was found
        """
        # Ensure images are in Pillow format
        image = to_image(image)
        template = to_image(template)

        # Crop image if requested
        if region is not None:
            region = to_region(region)
            image = image.crop(region.as_tuple())

        # Verify template still fits in image
        if template.size[0] > image.size[0] or template.size[1] > image.size[1]:
            raise ValueError("Template is larger than search region")

        # Strip alpha channels
        if image.mode == "RGBA":
            image = image.convert("RGB")
        if template.mode == "RGBA":
            template = template.convert("RGB")

        # Do the actual search
        start = time.time()
        matches = self.matcher.match(image, template, limit, tolerance)
        logging.info("Scanned image in %.2f seconds", time.time() - start)

        if not matches:
            raise ImageNotFoundError("No matches for given template")

        # Convert region coördinates back to full-size coördinates
        if region is not None:
            for match in matches:
                match.move(region.left, region.top)

        return matches

    def find_template_on_screen(self, template, **kwargs):
        """Attempt to find the template image from the current desktop.
        For argument descriptions, see ``find_template_in_image()``
        """
        return self.find_template_in_image(self.take_screenshot(), template, **kwargs)

    def wait_template_on_screen(self, template, timeout=5, **kwargs):
        """Wait for template image to appear on current desktop.
        For further argument descriptions, see ``find_template_in_image()``

        :param timeout: Time to wait for template (in seconds)
        """
        start_time = time.time()
        while time.time() - start_time > timeout:
            try:
                return self.find_template_on_screen(template, **kwargs)
            except ImageNotFoundError:
                time.sleep(0.1)

    def show_region_in_image(self, image, region, color="red", width=5):
        """Draw a rectangle onto the image around the given region.

        :param image:   image to draw onto
        :param region:  coordinates for region or Region object
        :param color:   color of rectangle
        :param width:   line width of rectangle
        """
        image = to_image(image)
        region = to_region(region)

        draw = ImageDraw.Draw(image)
        draw.rectangle(region.as_tuple(), outline=color, width=int(width))
        return image

    def show_region_on_screen(self, region, color="red", width=5):
        """Draw a rectangle around the given region on the current desktop.

        :param region:  coordinates for region or Region object
        :param color:   color of rectangle
        :param width:   line width of rectangle
        """
        image = self.take_screenshot()
        return self.show_region_in_image(image, region, color, width)

    def get_pixel_color_in_image(self, image, point):
        """Get the RGB value of a pixel in the image.

        :param image:   image to get pixel from
        :param point:   coordinates for pixel or Point object
        """
        point = to_point(point)
        pixel = image.getpixel(point.as_tuple())
        return RGB.from_pixel(pixel)

    def get_pixel_color_on_screen(self, point):
        """Get the RGB value of a pixel currently on screen.

        :param point:   coordinates for pixel or Point object
        """
        return self.get_pixel_color_in_image(self.take_screenshot(), point)


class TemplateMatcher:
    """Container class for different template matching methods."""

    DEFAULT_TOLERANCE = 0.95  # Tolerance for correlation matching methods

    def __init__(self, opencv=False):
        self.logger = logging.getLogger(__name__)
        self._opencv = opencv
        self._tolerance = self.DEFAULT_TOLERANCE
        self._tolerance_warned = False

    @property
    def tolerance(self):
        return self._tolerance

    @tolerance.setter
    def tolerance(self, value):
        self._tolerance = clamp(0.10, value, 1.00)

    def match(self, image, template, limit=None, tolerance=None):
        """Attempt to find the template in the given image.

        :param image:       image to search from
        :param template:    image to search with
        :param limit:       maximum number of returned matches
        :param tolerance:   minimum correlation factor between template and image
        :return:            list of regions that match criteria
        """
        if self._opencv:
            match_func = self._iter_match_opencv
        else:
            match_func = self._iter_match_pillow

        matches = []
        for match in match_func(image, template, tolerance):
            matches.append(match)
            if limit is not None and len(matches) >= int(limit):
                break

        return matches

    def _iter_match_opencv(self, image, template, tolerance):
        """Use opencv's matchTemplate() to slide the `template` over
        `image` to calculate correlation coefficients, and then
        filter with a tolerance to find all relevant global maximums.
        """
        # pylint: disable=no-member
        tolerance = tolerance or self.tolerance
        template_width, template_height = template.size

        image = numpy.array(image)
        template = numpy.array(template)

        image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
        template = cv2.cvtColor(template, cv2.COLOR_RGB2BGR)

        coefficients = cv2.matchTemplate(image, template, cv2.TM_CCOEFF_NORMED)

        while True:
            _, match_coeff, _, (match_x, match_y) = cv2.minMaxLoc(coefficients)
            if match_coeff < tolerance:
                break

            coefficients[
                match_y - template_height // 2 : match_y + template_height // 2,
                match_x - template_width // 2 : match_x + template_width // 2,
            ] = 0

            yield Region.from_size(match_x, match_y, template_width, template_height)

    def _iter_match_pillow(self, image, template, tolerance):
        """Brute-force search for template image in larger image.

        Use optimized string search for finding the first row and then
        check if whole template matches.

        TODO: Generalize string-search algorithm to work in two dimensions
        """
        if tolerance is not None and not self._tolerance_warned:
            self._tolerance_warned = True
            self.logger.warning(
                "Template matching tolerance not supported for current search method"
            )

        image = ImageOps.grayscale(image)
        template = ImageOps.grayscale(template)

        template_width, template_height = template.size
        template_rows = chunks(tuple(template.getdata()), template_width)

        image_width, _ = image.size
        image_rows = chunks(tuple(image.getdata()), image_width)

        for image_y, image_row in enumerate(image_rows[: -len(template_rows)]):
            for image_x in self._search_string(image_row, template_rows[0]):
                match = True
                for match_y, template_row in enumerate(template_rows[1:], image_y):
                    match_row = image_rows[match_y][image_x : image_x + template_width]
                    if template_row != match_row:
                        match = False
                        break

                if match:
                    yield Region.from_size(
                        image_x, image_y, template_width, template_height
                    )

    def _search_string(self, text, pattern):
        """Python implementation of Knuth-Morris-Pratt string search algorithm."""
        pattern_len = len(pattern)

        # Build table of shift amounts
        shifts = [1] * (pattern_len + 1)
        shift = 1
        for idx in range(pattern_len):
            while shift <= idx and pattern[idx] != pattern[idx - shift]:
                shift += shifts[idx - shift]
            shifts[idx + 1] = shift

        # Do the actual search
        start_idx = 0
        match_len = 0
        for char in text:
            while (
                match_len == pattern_len
                or match_len >= 0
                and pattern[match_len] != char
            ):
                start_idx += shifts[match_len]
                match_len -= shifts[match_len]
            match_len += 1
            if match_len == pattern_len:
                yield start_idx



import json
import logging
import os
import time
from typing import Any
import requests

from RPA.Robocloud.Secrets import Secrets

DEFAULT_REGION = "northeurope"


class AzureBase:
    """Base class for all Azure servives.

    `TOKEN_LIFESPAN` is in seconds, token is valid for 10 minutes so max lifetime
    is set to 9.5 minutes = 570.0 seconds
    """

    __base_url: str = None
    __region: str = None
    COGNITIVE_API = "api.cognitive.microsoft.com"
    TOKEN_LIFESPAN = 570.0
    region = None
    robocloud_vault_name: str = None
    services: dict = {}
    token = None
    token_time = None
    logger = None

    def _azure_request(
        self,
        service_name,
        url,
        method: str = "POST",
        token: str = None,
        params=None,
        content_type="application/json",
        filepath=None,
        jsondata=None,
        headers=None,
        body=None,
    ):
        if headers:
            request_parameters = {"headers": headers}
        else:
            request_parameters = {"headers": {}}
            if token:
                request_parameters["headers"]["Authorization"] = f"Bearer {token}"
            else:
                if (
                    service_name not in self.services
                    or self.services[service_name] is None
                ):
                    raise KeyError(
                        "Missing subscription key for service: %s" % service_name
                    )
                request_parameters["headers"][
                    "Ocp-Apim-Subscription-Key"
                ] = self.services[service_name]

            request_parameters["headers"]["Content-Type"] = content_type
        if filepath:
            with open(filepath, "rb") as f:
                filedata = f.read()
            request_parameters["data"] = filedata
        if jsondata:
            request_parameters["json"] = jsondata
        if params:
            request_parameters["params"] = params
        if body:
            request_parameters["data"] = body
        self.logger.debug("Azure %s: %s", method, url)
        self.logger.debug(
            "Content-Type: %s", request_parameters["headers"]["Content-Type"]
        )
        if method == "POST":
            response = requests.post(url, **request_parameters)
        elif method == "GET":
            response = requests.get(url, **request_parameters)
        else:
            raise KeyError("Unknown Azure request type: %s" % method)

        status_code = response.status_code
        self.logger.debug("Azure request response status: %s", status_code)
        assert (
            status_code == 200
        ), f"Azure responded with status: {status_code} '{response.reason}'"
        return response

    def _write_json(self, json_filepath, response_json):
        if json_filepath and response_json:
            with open(json_filepath, "w") as f:
                json.dump(response_json, f)

    def _set_subscription_key(self, service_name, use_robocloud_vault):
        common_key = "AZURE_SUBSCRIPTION_KEY"
        service_key = f"AZURE_{service_name.upper()}_KEY"
        sub_key = None
        if use_robocloud_vault:
            vault = Secrets()
            vault_items = vault.get_secret(self.robocloud_vault_name)
            vault_items = {k.upper(): v for (k, v) in vault_items.items()}
            if service_key in vault_items and vault_items[service_key].strip() != "":
                sub_key = vault_items[service_key]
            elif common_key in vault_items and vault_items[common_key].strip() != "":
                sub_key = vault_items[common_key]
            if sub_key is None:
                raise KeyError(
                    "The 'robocloud_vault_name' is required to access "
                    "Robocloud Vault. Set them in library "
                    "init or with `set_robocloud_vault` keyword."
                )
        else:
            sub_key = os.getenv(service_key)
            if sub_key is None or sub_key.strip() == "":
                sub_key = os.getenv(common_key)
            if sub_key is None or sub_key.strip() == "":
                raise KeyError(
                    "Azure service key is required to use Azure Cloud "
                    "service: %s" % service_name
                )

        self.services[service_name] = sub_key

    def _get_token(self, service_name, region):
        current_time = time.time()
        token_age = (current_time - self.token_time) if self.token_time else None
        if self.token is None or token_age is None or token_age > self.TOKEN_LIFESPAN:
            self.token = self._request_new_token(service_name, region)
            self.token_time = current_time
        return self.token

    def _request_new_token(self, service_name: str, region: str = None):
        self.logger.debug("Requesting new Azure token for accessing services")
        request_region = region if region else self.region
        self.__base_url = f"https://{request_region}.{self.COGNITIVE_API}"
        token_url = f"{self.__base_url}/sts/v1.0/issueToken"
        response = self._azure_request(service_name, token_url)
        return str(response.text)

    def _handle_response(self, json_filepath: None, response: None):
        response_json = response.json()
        self._write_json(json_filepath, response_json)
        return response_json

    def _azure_request_image(
        self, service_name, request_url, image_url, image_file, params=None
    ):
        if image_url:
            return self._azure_request(
                service_name,
                request_url,
                params=params,
                jsondata={"url": image_url},
                content_type="application/json",
            )
        else:
            return self._azure_request(
                service_name,
                request_url,
                params=params,
                filepath=image_file,
                content_type="application/octet-stream",
            )

    def _image_url_or_image_file_is_required(
        self, image_url: str = None, image_file: str = None
    ):
        if image_url is None and image_file is None:
            raise KeyError("Parameter 'image_url' or 'image_file' must be given.")

    def set_robocloud_vault(self, vault_name):
        """Set Robocloud Vault name

        :param vault_name: Robocloud Vault name
        """
        if vault_name:
            self.robocloud_vault_name = vault_name


class ServiceTextAnalytics(AzureBase):
    """Class for Azure TextAnalytics service"""

    __service_name = "textanalytics"

    def __init__(self) -> None:
        self.logger.debug("ServiceTextAnalytics init")
        self.__region = None

    def init_text_analytics_service(
        self, region: str = None, use_robocloud_vault: bool = False
    ):
        """Initialize Azure Text Analyticts

        :param region: identifier for service region
        :param use_robocloud_vault: use secret stored into `Robocloud Vault`
        """
        self.__region = region if region else self.region
        self.__base_url = f"https://{self.__region}.{self.COGNITIVE_API}"
        self._set_subscription_key(self.__service_name, use_robocloud_vault)

    def sentiment_analyze(
        self, text: str, language: str = None, json_file: str = None
    ) -> dict:
        """Analyze sentiments in the given text

        :param text: A UTF-8 text string
        :param language: if input language is known
        :param json_file: filepath to write results into
        :return: analysis in json format
        """
        analyze_url = f"{self.__base_url}/text/analytics/v3.0/sentiment"

        document = {"id": "1", "text": text}
        if language:
            document["language"] = language
        response = self._azure_request(
            self.__service_name, analyze_url, jsondata={"documents": [document]}
        )
        return self._handle_response(json_file, response)

    def detect_language(self, text: str, json_file: str = None) -> dict:
        """Detect languages in the given text

        :param text: A UTF-8 text string
        :param json_file: filepath to write results into
        :return: analysis in json format
        """
        analyze_url = f"{self.__base_url}/text/analytics/v3.0/languages"
        document = {"id": "1", "text": text}
        response = self._azure_request(
            self.__service_name, analyze_url, jsondata={"documents": [document]}
        )
        return self._handle_response(json_file, response)

    def key_phrases(
        self, text: str, language: str = None, json_file: str = None
    ) -> dict:
        """Detect key phrases in the given text

        :param text: A UTF-8 text string
        :param language: if input language is known
        :param json_file: filepath to write results into
        :return: analysis in json format
        """
        analyze_url = f"{self.__base_url}/text/analytics/v3.0/keyphrases"
        document = {"id": "1", "text": text}
        if language:
            document["language"] = language
        response = self._azure_request(
            self.__service_name, analyze_url, jsondata={"documents": [document]}
        )
        return self._handle_response(json_file, response)

    def find_entities(self, text: str, language: str = None, json_file=None) -> dict:
        """Detect entities in the given text

        :param text: A UTF-8 text string
        :param language: if input language is known
        :param json_file: filepath to write results into
        :return: analysis in json format
        """
        analyze_url = f"{self.__base_url}/text/analytics/v2.1/entities"
        document = {"id": "1", "text": text}
        if language:
            document["language"] = language
        response = self._azure_request(
            self.__service_name, analyze_url, jsondata={"documents": [document]}
        )
        return self._handle_response(json_file, response)


class ServiceFace(AzureBase):
    """Class for Azure Face service"""

    __service_name = "face"

    def __init__(self) -> None:
        self.logger.debug("ServiceFace init")

    def init_face_service(
        self, region: str = None, use_robocloud_vault: bool = False
    ) -> None:
        """Initialize Azure Face

        :param region: identifier for service region
        :param use_robocloud_vault: use secret stored into `Robocloud Vault`
        """
        self.__region = region if region else self.region
        self.__base_url = f"https://{self.__region}.{self.COGNITIVE_API}"
        self._set_subscription_key(self.__service_name, use_robocloud_vault)

    def detect_face(
        self,
        image_file: str = None,
        image_url: str = None,
        face_attributes: str = None,
        face_landmarks: bool = False,
        recognition_model: str = "recognition_02",
        json_file: str = None,
    ) -> dict:
        # pylint: disable=C0301
        """Detect facial attributes in the image

        Read more about `face_attributes` at `Face detection explained`_:

            - age
            - gender
            - smile
            - facialHair
            - headPose
            - glasses
            - emotion
            - hair
            - makeup
            - accessories
            - blur
            - exposure
            - nouse

        :param image_file: filepath of image file
        :param image_url: URI to image, if given will be used instead of `image_file`
        :param face_attributes: comma separated list of attributes,
            for example. "age,gender,smile"
        :param face_landmarks: return face landmarks of the detected faces
            or not. The default value is `False`
        :param recognition_model: model used by Azure to detech faces, options
            are "recognition_01" or "recognition_02", default is "recognition_02"
        :param json_file: filepath to write results into
        :return: analysis in json format

        .. _Face detection explained: https://docs.microsoft.com/en-us/azure/cognitive-services/face/concepts/face-detection  # noqa: E501
        """
        self._image_url_or_image_file_is_required(image_url, image_file)
        analyze_url = f"{self.__base_url}/face/v1.0/detect"
        params = {
            "returnFaceId": "true",
            "returnFaceLandmarks": str(face_landmarks),
            "recognitionModel": recognition_model,
            "returnRecognitionModel": "false",
        }
        if face_attributes:
            params["returnFaceAttributes"] = face_attributes
        response = self._azure_request_image(
            self.__service_name, analyze_url, image_url, image_file, params
        )
        return self._handle_response(json_file, response)


class ServiceComputerVision(AzureBase):
    """Class for Azure Computer Vision service"""

    __service_name = "computervision"

    def __init__(self) -> None:
        self.logger.debug("ServiceComputerVision init")

    def init_computer_vision_service(
        self, region: str = None, use_robocloud_vault: bool = False
    ) -> None:
        """Initialize Azure Computer Vision

        :param region: identifier for service region
        :param use_robocloud_vault: use secret stored into `Robocloud Vault`
        """
        self.__region = region if region else self.region
        self.__base_url = f"https://{self.__region}.{self.COGNITIVE_API}"
        self._set_subscription_key(self.__service_name, use_robocloud_vault)

    def vision_analyze(
        self,
        image_file: str = None,
        image_url: str = None,
        visual_features: str = None,
        json_file: str = None,
    ) -> dict:
        # pylint: disable=C0301
        """Identify features in the image

        See `Computer Vision API`_ for valid feature names and their explanations:

            - Adult
            - Brands
            - Categories
            - Color
            - Description
            - Faces
            - ImageType
            - Objects
            - Tags

        :param image_file: filepath of image file
        :param image_url: URI to image, if given will be used instead of `image_file`
        :param visual_features: comma separated list of features,
            for example. "Categories,Description,Color"
        :param json_file: filepath to write results into
        :return: analysis in json format

        .. _Computer Vision API: https://westcentralus.dev.cognitive.microsoft.com/docs/services/computer-vision-v3-ga  # noqa: E501
        """
        self._image_url_or_image_file_is_required(image_url, image_file)
        analyze_url = f"{self.__base_url}/vision/v3.0/analyze"
        params = {}
        if visual_features:
            params["visualFeatures"] = visual_features
        response = self._azure_request_image(
            self.__service_name, analyze_url, image_url, image_file, params
        )
        return self._handle_response(json_file, response)

    def vision_describe(
        self, image_file: str = None, image_url: str = None, json_file: str = None
    ) -> dict:
        """Describe image with tags and captions

        :param image_file: filepath of image file
        :param image_url: URI to image, if given will be used instead of `image_file`
        :param json_file: filepath to write results into
        :return: analysis in json format
        """
        self._image_url_or_image_file_is_required(image_url, image_file)
        analyze_url = f"{self.__base_url}/vision/v3.0/describe"
        params = {"maxCandidates": "3", "language": "en"}
        response = self._azure_request_image(
            self.__service_name, analyze_url, image_url, image_file, params
        )
        return self._handle_response(json_file, response)

    def vision_ocr(
        self, image_file: str = None, image_url: str = None, json_file: str = None
    ) -> dict:
        """Optical Character Recognition (OCR) detects text in an image

        :param image_file: filepath of image file
        :param image_url: URI to image, if given will be used instead of `image_file`
        :param json_file: filepath to write results into
        :return: analysis in json format
        """
        self._image_url_or_image_file_is_required(image_url, image_file)
        analyze_url = f"{self.__base_url}/vision/v3.0/ocr"
        params = {"detectOrientation": "true", "language": "en"}
        response = self._azure_request_image(
            self.__service_name, analyze_url, image_url, image_file, params
        )
        return self._handle_response(json_file, response)

    def vision_detect_objects(
        self, image_file: str = None, image_url: str = None, json_file: str = None
    ) -> dict:
        """Detect objects in the image

        :param image_file: filepath of image file
        :param image_url: URI to image, if given will be used instead of `image_file`
        :param json_file: filepath to write results into
        :return: analysis in json format
        """
        self._image_url_or_image_file_is_required(image_url, image_file)
        analyze_url = f"{self.__base_url}/vision/v3.0/detect"
        response = self._azure_request_image(
            self.__service_name, analyze_url, image_url, image_file
        )
        return self._handle_response(json_file, response)


class ServiceSpeech(AzureBase):
    """Class for Azure Speech service"""

    __service_name = "speech"
    audio_formats = {
        "MP3": "audio-24khz-96kbitrate-mono-mp3",
        "WAV": "riff-24khz-16bit-mono-pcm",
    }

    def __init__(self) -> None:
        self.logger.debug("ServiceSpeech init")

    def init_speech_service(
        self, region: str = None, use_robocloud_vault: bool = False
    ) -> None:
        """Initialize Azure Speech

        :param region: identifier for service region
        :param use_robocloud_vault: use secret stored into `Robocloud Vault`
        """
        self.__region = region if region else self.region
        self.__base_url = f"https://{self.__region}.tts.speech.microsoft.com"
        self._set_subscription_key(self.__service_name, use_robocloud_vault)

    def text_to_speech(
        self,
        text: str,
        language: str = "en-US",
        name: str = "en-US-AriaRUS",
        gender: str = "FEMALE",
        encoding: str = "MP3",
        neural_voice_style: Any = None,
        target_file: str = "synthesized.mp3",
    ):
        """Synthesize speech synchronously

        Neural voices are only supported for Speech resources created in
        East US, South East Asia, and West Europe regions.

        :param text: input text to synthesize
        :param language: voice language, defaults to "en-US"
        :param name: voice name, defaults to "en-US-AriaRUS"
        :param gender: voice gender, defaults to "FEMALE"
        :param encoding: result encoding type, defaults to "MP3"
        :param neural_voice_style: if given then neural voice is used,
            example style. "cheerful"
        :param target_file: save synthesized output to file,
            defaults to "synthesized.mp3"
        :return: synthesized output in bytes
        """
        encoding = encoding.upper() if encoding else None
        if encoding is None or encoding not in self.audio_formats.keys():
            raise KeyError(
                "Unknown encoding %s for text_to_speech, available formats are: %s"
                % (encoding, ", ".join(self.audio_formats.keys()))
            )
        token = self._get_token(self.__service_name, self.__region)
        headers = {
            "X-Microsoft-OutputFormat": self.audio_formats[encoding],
            "Content-Type": "application/ssml+xml",
            "Authorization": "Bearer " + token,
            "User-Agent": "RPA software robot",
        }
        voice_url = f"{self.__base_url}/cognitiveservices/v1"
        xml_ns = "xmlns='http://www.w3.org/2001/10/synthesis'"
        if neural_voice_style is None:
            request_body = f"""
            <speak version='1.0' xml:lang='en-US'>
                <voice xml:lang='{language}'
                    xml:gender='{gender.capitalize()}'
                    name='{name}'>
            {text}
            </voice></speak>
            """
        else:
            mstts_ns = "xmlns:mstts='https://www.w3.org/2001/mstts'"
            request_body = f"""
            <speak version="1.0" {xml_ns} {mstts_ns} xml:lang="en-US">
                <voice name="{name}">
                    <mstts:express-as style="{neural_voice_style}">
                    {text}
                    </mstts:express-as>
                </voice>
            </speak>"""

        response = self._azure_request(
            self.__service_name,
            voice_url,
            method="POST",
            token=token,
            headers=headers,
            body=request_body,
        )
        if target_file:
            with open(target_file, "wb") as f:
                f.write(response.content)
        return response.content

    def list_supported_voices(
        self, locale: str = None, neural_only: bool = False, json_file: str = None
    ):
        """List supported voices for Azure API Speech Services.

        Available voice selection might differ between regions.

        :param locale: list only voices specific to locale, by default return all voices
        :param neural_only: `True` if only neural voices should be returned,
            `False` by default
        :param json_file: filepath to write results into
        :return: voices in json
        """
        token = self._get_token(self.__service_name, self.__region)
        voice_url = f"{self.__base_url}/cognitiveservices/voices/list"
        response = self._azure_request(
            self.__service_name, voice_url, method="GET", token=token
        )
        response_json = response.json()
        if neural_only:
            response_json = [x for x in response_json if x["VoiceType"] == "Neural"]
        if locale:
            response_json = [x for x in response_json if locale in x["Locale"]]
        self._write_json(json_file, response_json)
        return response_json


class Azure(ServiceTextAnalytics, ServiceFace, ServiceComputerVision, ServiceSpeech):
    """Library for interacting with Azure services

    Authentication to Azure API is handled by `service subcription key` which
    can set for all services using environment variable `AZURE_SUBSCRIPTION_KEY`.

    If there are service specific subscription keys, these can be set using
    environment variable `AZURE_SERVICENAME_KEY`. Replace `SERVICENAME` with service
    name.
    """

    def __init__(self, region: str = DEFAULT_REGION, robocloud_vault_name: str = None):
        self.set_robocloud_vault(robocloud_vault_name)
        self.logger = logging.getLogger(__name__)
        ServiceTextAnalytics.__init__(self)
        ServiceFace.__init__(self)
        ServiceComputerVision.__init__(self)
        ServiceSpeech.__init__(self)
        self.region = region
        self.logger.info("Azure library initialized. Default region: %s", self.region)



import base64
import copy
import json
import logging
import re
import string
import sys
from functools import partial
from pathlib import Path

import graphviz
from graphviz import ExecutableNotFound
from robot.errors import PassExecution
from robot.libraries.BuiltIn import BuiltIn


class SchemaError(Exception):
    """Error raised when violating schema."""


class Graph:
    """Task execution graph.

    Helper class which is used to store executed tasks and
    transitions between them, and render the relationships
    in a digraph.

    Creates a dot-notation file and rendered graph using graphviz.

    :param suite: Current suite running model
    """

    # Default render format
    FORMAT = "svg"

    # Default attributes
    GRAPH = {
        "rankdir": "LR",
    }
    NODE = {
        "shape": "box",
        "style": "rounded,filled",
        "margin": "0.15,0.1",
        "height": "0",
        "fontname": "Helvetica, Arial, sans-serif",
        "fontsize": "12",
    }

    # Start / end nodes
    TERMINATOR = {
        "shape": "oval",
        "style": "filled",
        "color": "#cccccc",
        "fillcolor": "#eeeeee",
        "fontcolor": "#0f0f0f",
    }

    # Node background/font colors
    COLORS = {
        "none": {"color": "#eeeeee", "fontcolor": "#0f0f0f"},
        "fail": {"color": "#d9534f", "fontcolor": "#ffffff"},
        "pass": {"color": "#5cb85c", "fontcolor": "#ffffff"},
        "warn": {"color": "#ec971f", "fontcolor": "#ffffff"},
    }

    def __init__(self, suite):
        #: Current suite
        self.suite = suite
        #: Task data by name
        self.tasks = {}
        #: Transition pairs between tasks
        self.edges = set()
        #: Current running task
        self.current = None
        #: Flag for successful end of process
        self.is_end = False

        self._parse_tasks(suite.tests)

    def _parse_tasks(self, tasks):
        """Parse tasks (nodes) and assign unique labels."""
        for position, task in enumerate(tasks):
            label = self._create_label(position)
            self.tasks[task.name] = {
                "name": task.name,
                "label": label,
                "result": "none",
                "doc": task.doc or task.name,
            }

    @staticmethod
    def _create_label(position):
        """Generate label for node, e.g. A, B, ..., AA, AB, AC, ..."""
        letters = string.ascii_uppercase

        label = ""
        while True:
            position, index = divmod(position, len(letters))
            label = letters[index] + label
            if not position:
                return label
            position -= 1

    def _create_graph(self, strip=True):
        """Create graphviz graph from current execution state."""
        graph = graphviz.Digraph(
            name=self.suite.name,
            format=self.FORMAT,
            graph_attr=self.GRAPH,
            node_attr=self.NODE,
        )

        # Start/end nodes
        graph.node("Start", **self.TERMINATOR)
        if self.is_end:
            graph.node("End", **self.TERMINATOR)

        # Task nodes
        for task in self.tasks.values():
            result = task.get("result", "none")
            if not (result == "none" and strip):
                colors = self.COLORS[result]
                graph.node(task["label"], task["name"], tooltip=task["doc"], **colors)

        # Edges
        for src, dst in self.edges:
            src = src if src == "Start" else self.tasks[src]["label"]
            dst = dst if dst == "End" else self.tasks[dst]["label"]
            graph.edge(src, dst)

        return graph

    def render_to_file(self, path, strip=True):
        """Render graphviz graph to given file."""
        path = Path(path)
        graph = self._create_graph(strip)
        return graph.render(filename=path.name, directory=path.parent)

    def render_to_bytes(self, strip=True):
        """Render graphviz graph to in-memory bytes object."""
        graph = self._create_graph(strip)
        return graph.pipe()

    def set_next(self, task):
        """Add transition between previous and next task."""
        previous, self.current = self.current, task.name
        if not previous:
            assert not self.edges, "Edges exist without previous task"
            previous = "Start"

        assert not self.is_end, f"Attempting to add task after end: {self.current}"
        assert self.current in self.tasks, f"Unknown task: {self.current}"

        pair = (previous, self.current)
        if pair not in self.edges:
            self.edges.add(pair)

    def set_result(self, result):
        """Set execution result for current task."""
        assert not self.is_end, "End already set"
        task = self.tasks[self.current]
        task["result"] = str(result).lower()

    def set_end(self):
        """Add final edge to End node."""
        assert not self.is_end, "End already set"
        self.edges.add((self.current, "End"))
        self.is_end = True


class Schema:
    """Task execution schema.

    A library for validating transitions betweens tasks,
    and evaluating possible schema-defined actions when
    these transitions are triggered.

    :param schema: content of schema JSON file
    :param names:  names of tasks in the current suite
    """

    def __init__(self, schema, names):
        #: Schema properties by task name
        self.tasks = {}
        #: Aliases for tasks
        self.aliases = {}
        #: First task in execution
        self.start = None
        #: Allowed end task(s)
        self.end = []

        self._parse_schema(schema, names)

    def _parse_schema(self, schema, names):
        """Parse schema file and validate contents.

        :param schema: content of schema file
        :param tasks:  tasks in suite execution
        """
        # First pass: Parse names and aliases
        for name, properties in schema.get("tasks", {}).items():
            assert name in names, f"Unknown task name: {name}"
            assert name not in self.tasks, f"Duplicate task name: {name}"

            # Flag for first task in the execution
            if properties.get("start", False):
                assert self.start is None, "Duplicate start task"
                self.start = name

            # Flag for allowed end task
            if properties.get("end", False):
                self.end.append(name)

            # Optional task alias
            alias = properties.get("alias")
            if alias:
                assert alias not in self.aliases, f"Duplicate alias: {alias}"
                self.aliases[alias] = name

            self.tasks[name] = properties

        # Second pass: Parse references to other tasks
        for name, properties in self.tasks.items():
            # Whitelist of allowed next tasks
            if "next" in properties:
                properties["next"] = [
                    self.resolve_reference(task) for task in properties["next"]
                ]
            # Actions for resolving the next task
            if "actions" in properties:
                properties["actions"] = [
                    self._create_action(action) for action in properties["actions"]
                ]

        # No start defined in schema, fallback to first in suite
        if not self.start:
            self.start = names[0]

    def _create_action(self, action):
        """Convert action definition in schema to callable."""
        assert "task" in action, "Next task undefined for action"
        task = self.resolve_reference(action["task"])

        callbacks = {
            "exception": self._action_exception,
            "condition": self._action_condition,
            "status": self._action_status,
        }

        operator = set(callbacks) & set(action)
        assert operator, f"Unknown action definition: {action}"
        assert len(operator) == 1, f"Multiple conditions in action: {action}"

        operator = operator.pop()
        callback = callbacks[operator]
        callback = partial(callback, action[operator])

        return callback, task

    def _action_exception(self, pattern, result):
        """Schema action: catch exception if it matches message pattern."""
        if result.passed:
            return False

        if not re.match(pattern, result.message):
            return False

        result.message = f"Transition: message = {pattern}"
        result.status = "PASS"
        return True

    def _action_condition(self, condition, result):
        """Schema action: evaluate Robot Framework expression."""
        if not result.passed and result.critical:
            return False

        if not BuiltIn().evaluate(condition):
            return False

        result.message = f"Transition: {condition}"
        return True

    def _action_status(self, status, result):
        """Schema action: compare test result to expected."""
        if result.status.upper() != status.upper():
            return False

        result.message = f"Transition: status == {status}"
        return True

    def resolve_reference(self, name):
        """Convert task reference to original name."""
        if name in self.tasks:
            return name
        elif name in self.aliases:
            return self.aliases[name]
        else:
            raise ValueError(f"Unknown task or alias: {name}")

    def validate(self, src, dst):
        """Validate transition between two tasks."""
        assert src in self.tasks, f"Unknown source task: {src}"

        # Optional end task validation
        if dst == "end":
            if self.end and src not in self.end:
                raise SchemaError("Unexpected end task")
            return

        if dst not in self.tasks:
            raise SchemaError(f"Destination '{dst}' not in schema")

        if "next" in self.tasks[src] and dst not in self.tasks[src]["next"]:
            raise SchemaError(f"Invalid transition '{src}' -> '{dst}'")

    def evaluate_actions(self, src, result):
        """Evaluate all actions for the source task,
        and return any potential triggered destination tasks.
        """
        actions = self.tasks[src].get("actions", [])

        # Evaluate all callbacks
        for action, task in actions:
            if action(result):
                return task

        # No conditions matched
        return None


class Tasks:
    """`Tasks` is a Robot Framework library for controlling
    task execution inside a suite. It allows changing the next scheduled task
    or jumping immediately to other tasks by using the keywords it provides.

    If ``Graphviz`` is available, it will also create a directed graph
    to visualize connected tasks, which is visible in the suite documentation
    field of the test log.

    :param execution_limit: Maximum number of tasks to run in suite,
                            used to prevent infinite loops
    :param schema:          Path to optional schema file
    :param graph:           Render execution result as graph using graphviz
    :param graph_inline:    Inline graph into log, instead of saving as file
    """

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LISTENER_API_VERSION = 3

    TAG_NONCRITICAL = "tasks-schema-noncritical"

    def __init__(
        self, execution_limit=1024, schema=None, graph=True, graph_inline=True
    ):
        self.ROBOT_LIBRARY_LISTENER = self

        #: Current task execution count
        self.count = 0
        #: Maximum task execution
        self.limit = int(execution_limit)
        #: Current suite running model
        self.suite = None
        #: Original task running models
        self.tasks = None
        #: Current running task
        self.current = None
        #: Next scheduled task
        self.next = None
        #: Task execution schema
        self.schema = None
        self.schema_path = schema
        #: Task execution graph
        self.graph = None
        self.graph_options = {"enabled": graph, "inline": graph_inline}

    def _load_schema(self):
        """Load schema from file, if defined."""
        self.schema = None
        if not self.schema_path:
            return

        with open(self.schema_path) as fd:
            data = json.load(fd)
            names = [task.name for task in self.tasks]
            self.schema = Schema(data, names)

    def _task_by_name(self, name):
        """Find task execution object by shortname."""
        for task in self.tasks:
            if task.name == name:
                return task

        raise ValueError(f"Task not found: {name}")

    def _find_next_task(self, result):
        """Resolve the next task based on the previous result."""
        # TODO: Move all result object modifying here
        task = None

        try:
            if self.next:
                task = self.next
                result.message = "Transition: Set by keyword"
            elif self.schema:
                task = self.schema.evaluate_actions(self.current.name, result)
                task = self._task_by_name(task) if task else None
                result.tags.add(self.TAG_NONCRITICAL)  # Schema overrides status

            if self.schema:
                name = task.name if task else "end"
                self.schema.validate(self.current.name, name)

        except SchemaError as err:
            logging.error(err)
            result.status = "FAIL"

        finally:
            self.next = None

        return task, result

    def _append_task(self, task):
        """Append new copy of task to execution queue."""
        self.count += 1

        # Ensure we don't edit original model
        name = "#{:<3} {}".format(self.count, task.name)
        copied = task.copy(name=name)
        self.suite.tests.append(copied)
        self.current = task

        # Show transition between tasks
        self.graph.set_next(task)

    def _start_suite(self, data, result):
        """Robot listener method, called on suite start.
        Copies original tasks to be used as source for scheduling.
        """
        self.count = 0
        self.next = None
        self.suite = data
        self.tasks = copy.deepcopy(self.suite.tests)
        self.graph = Graph(self.suite)

        self.suite.tests.clear()

        try:
            self._load_schema()
        except Exception as exc:  # pylint: disable=broad-except
            logging.error("Schema parsing failed: %s", exc)
            sys.exit(1)

        if self.schema:
            self._append_task(self._task_by_name(self.schema.start))
        else:
            self._append_task(self.tasks[0])

        # Set a unique non-critical tag to be used when
        # execution should continue after a FAIL status
        root = result
        while root.parent:
            root = root.parent
        root.set_criticality(non_critical_tags=[self.TAG_NONCRITICAL])

    def _end_suite(self, data, result):
        """Render graph of suite execution to the documentation field."""
        del result

        if not self.graph_options.get("enabled", True):
            return

        try:
            if self.graph_options.get("inline", True):
                # Render as inline data URI
                data = self.graph.render_to_bytes()
                src = "data:image/svg+xml;base64,{}".format(
                    base64.b64encode(data).decode("utf-8")
                )
            else:
                # Render to file
                dirname = BuiltIn().get_variable_value("${OUTPUT_DIR}")
                filename = "graph_{}".format(data.name.lower().replace(" ", "_"))

                path = Path(dirname, filename)
                src = self.graph.render_to_file(str(path))

            BuiltIn().set_suite_documentation(f"[{src}|Graph]", append=True)
        except ExecutableNotFound as err:
            logging.warning("Graphviz executable not found: %s", err)

    def _end_test(self, data, result):
        """Robot listener method, called on test end.
        Rewrites next executable task, if overriden by keywords.
        Appends incrementing number to prevent task naming conflicts.
        """
        del data

        task, result = self._find_next_task(result)

        if not result.passed:
            self.graph.set_result("fail")
            if result.critical:
                return
        else:
            self.graph.set_result("pass")

        if not task:
            self.graph.set_end()
            return
        else:
            self._append_task(task)

    def set_next_task(self, name):
        """Set the next task to be executed.
        Should be a task in the same suite.

        :param name: Name of next task
        """
        task = self._task_by_name(name)

        if self.next:
            logging.warning(
                "Overwriting scheduled task '%s' with '%s'", self.next.name, task.name
            )

        logging.info("Scheduling task: %s", task.name)
        assert self.count < self.limit, "Reached task execution limit"

        if self.schema:
            self.schema.validate(self.current.name, task.name)

        self.next = task

    def set_next_task_if(self, condition, name, default=None):
        """Set the next task according to the condition.
        If no default is given, does not modify execution order.

        :param condition: Condition expression to evaluate
        :param name:      Name of next task, if successful
        :param default:   Name of next task, if unsuccessful
        """
        is_true = (
            BuiltIn().evaluate(condition)
            if isinstance(condition, str)
            else bool(condition)
        )

        logging.info("Condition: %s -> %s", condition, is_true)
        task = name if is_true else default

        if task:
            self.set_next_task(task)

    def jump_to_task(self, name):
        """Jump directly to given task, skipping the rest of the task
        execution. If run inside a teardown, also skips the rest of the
        teardown sequence.
        """
        self.set_next_task(name)
        raise PassExecution(f"Jumping to: {self.next}")

    def jump_to_task_if(self, condition, name, default=None):
        """Jump directly to given task according to the condition."""
        self.set_next_task_if(condition, name, default)
        if self.next:
            raise PassExecution(f"Jumping to: {self.next}")

    def set_next_task_if_keyword_fails(self, task, keyword, *args):
        """Executes given keyword and sets the next task if it fails."""
        success = BuiltIn().run_keyword_and_return_status(keyword, *args)
        if not success:
            self.set_next_task(task)

    def set_next_task_if_keyword_succeeds(self, task, keyword, *args):
        """Executes given keyword and sets the next task if it succeeds."""
        success = BuiltIn().run_keyword_and_return_status(keyword, *args)
        if success:
            self.set_next_task(task)

    def jump_to_task_if_keyword_fails(self, task, keyword, *args):
        """Executes given keyword and jumps to given task if it fails."""
        success = BuiltIn().run_keyword_and_return_status(keyword, *args)
        if not success:
            self.jump_to_task(task)

    def jump_to_task_if_keyword_succeeds(self, task, keyword, *args):
        """Executes given keyword and jumps to given task if it succeeds."""
        success = BuiltIn().run_keyword_and_return_status(keyword, *args)
        if success:
            self.jump_to_task(task)



import base64
import importlib
import logging
import os
import platform
import time
import traceback
from functools import partial
from itertools import product
from typing import Any, Optional
from pathlib import Path

from robot.libraries.BuiltIn import BuiltIn, RobotNotRunningError
from SeleniumLibrary import SeleniumLibrary, EMBED
from SeleniumLibrary.base import keyword
from SeleniumLibrary.keywords import BrowserManagementKeywords, ScreenshotKeywords
from selenium.webdriver import ChromeOptions

from RPA.core import locators, notebook, webdriver


def html_table(header, rows):
    """Create HTML table that can be used for logging."""
    output = '<div class="doc"><table>'
    output += "<tr>" + "".join(f"<th>{name}</th>" for name in header) + "</tr>"
    for row in rows:
        output += "<tr>" + "".join(f"<td>{name}</td>" for name in row) + "</tr>"
    output += "</table></div>"
    return output


class BrowserNotFoundError(ValueError):
    """Raised when browser can't be initialized."""


class Browser(SeleniumLibrary):
    """RPA Framework library for Browser operations.

    Extends functionality of SeleniumLibrary, for more information see
    https://robotframework.org/SeleniumLibrary/SeleniumLibrary.html
    """

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    AUTOMATIC_BROWSER_SELECTION = "AUTO"

    AVAILABLE_OPTIONS = {
        "chrome": "ChromeOptions",
        "firefox": "FirefoxOptions",
        # "safari": "WebKitGTKOptions",
        # "ie": "IeOptions",
    }

    def __init__(self, *args, **kwargs) -> None:
        locators_path = kwargs.pop("locators_path", locators.default_locators_path())

        # Parse user-given plugins
        plugins = kwargs.get("plugins", "")
        plugins = set(p for p in plugins.split(",") if p)

        # Add testability if requested
        if "use_testability" in args:
            args = [arg for arg in args if arg != "use_testability"]
            plugins.add("SeleniumTestability")

        # Refresh plugins list
        kwargs["plugins"] = ",".join(plugins)

        SeleniumLibrary.__init__(self, *args, **kwargs)
        self.logger = logging.getLogger(__name__)
        self.using_testability = bool("SeleniumTestability" in plugins)

        # Add support for locator aliases
        self.locators = locators.LocatorsDatabase(locators_path)
        self._element_finder.register("alias", self._find_by_alias, persist=True)

        self._embedding_screenshots = False
        self._previous_screenshot_directory = None
        # Embed screenshots in logs by default
        if not notebook.IPYTHON_AVAILABLE:
            self._embedding_screenshots = True
            self._previous_screenshot_directory = self.set_screenshot_directory(EMBED)

    @property
    def location(self) -> str:
        """Return browser location."""
        return self.get_location()

    def _find_by_alias(self, parent, criteria, tag, constraints):
        """Custom 'alias' locator that uses locators database."""
        del constraints

        if not Path(self.locators.path).exists():
            self.logger.warning("File does not exist: %s", self.locators.path)

        self.locators.load()
        if self.locators.error:
            error_msg, error_args = self.locators.error
            raise ValueError(error_msg % error_args)

        entry = self.locators.find_by_name(criteria)
        if not entry:
            raise ValueError(f"Unknown locator alias: {criteria}")

        if entry["type"] != "browser":
            raise ValueError(f"Not a browser locator: {criteria}")

        locator = "{prefix}:{criteria}".format(
            prefix=entry["strategy"], criteria=entry["value"]
        )

        self.logger.info("%s is an alias for %s", criteria, locator)
        return self._element_finder.find(locator, tag, parent)

    @keyword
    def open_available_browser(
        self,
        url: str,
        use_profile: bool = False,
        headless: Optional[bool] = None,
        maximized: bool = False,
        browser_selection: Any = AUTOMATIC_BROWSER_SELECTION,
        alias: Optional[str] = None,
        profile_name: Optional[str] = None,
        profile_path: Optional[str] = None,
        preferences: Optional[dict] = None,
        proxy: str = None,
    ) -> int:
        # pylint: disable=C0301
        """Opens the first available browser in the system in preferred order, or the
        given browser (``browser_selection``).

        ``url`` URL to open

        ``use_profile`` Set browser profile, default ``False`` (Chrome/Chromium only)

        ``headless`` Run in headless mode, default ``False``

        ``maximized`` Run window maximized, default ``False``

        ``browser_selection`` browser name, default ``AUTOMATIC_BROWSER_SELECTION``

        ``alias`` Custom name for browser

        ``profile_name`` Name of profile (if profile enabled)

        ``profile_path`` Path to profiles (if profile enabled)

        ``preferences`` Profile preferences (Chrome/Chromium only)

        ``proxy`` Proxy server address (Chrome only)

        Returns an index of the webdriver session.

        === Process of opening a browser ===

        1. Get the order of browsers

        2. Loop the list of preferred browsers

            a. Set the webdriver options for the browser

            b. Create the webdriver using existing installation

            c. (If step b. failed) Download and install webdriver, try again

            d. (If step c. failed) Try starting webdriver in headless mode

        3. Open the URL

        Returns index or custom alias for the browser instance.

        Raises ``BrowserNotFoundError`` if unable to open the browser.

        For information about Safari webdriver setup, see
        https://developer.apple.com/documentation/webkit/testing_with_webdriver_in_safari

        Example:
            | ${idx1} | Open Available Browser | https://www.robocorp.com |
            | ${idx2} | Open Available Browser | ${URL} | browser_selection=opera,firefox |
            | Open Available Browser | ${URL} | headless=True | proxy=localhost:8899 |
        """  # noqa: E501
        # pylint: disable=redefined-argument-from-local
        browser_options = self._get_browser_order(browser_selection)
        headless_options = [headless] if headless is not None else [False, True]
        download_options = [False, True]

        # Try a combination of all options until a browser starts
        index_or_alias = None
        options = []
        for browser, headless, download in product(
            browser_options, headless_options, download_options
        ):
            try:
                self.logger.debug(
                    "Creating webdriver for '%s' (headless: %s, download: %s)",
                    browser,
                    headless,
                    download,
                )
                kwargs, arguments = self._get_driver_args(
                    browser,
                    headless,
                    maximized,
                    use_profile,
                    profile_name,
                    profile_path,
                    preferences,
                    proxy,
                )
                index_or_alias = self._create_webdriver(
                    browser, alias, download, **kwargs
                )
                options.append((browser, headless, download, ""))
                self.logger.info(
                    "Created %s browser with arguments: %s",
                    browser,
                    " ".join(arguments),
                )
                break
            except Exception as error:  # pylint: disable=broad-except
                options.append((browser, headless, download, error))
                self.logger.debug(traceback.format_exc())

        # Log table of all attempted combinations
        try:
            table = html_table(
                header=["Browser", "Headless", "Download", "Error"], rows=options
            )
            BuiltIn().log("<p>Attempted combinations:</p>" + table, html=True)
        except RobotNotRunningError:
            pass

        # No webdriver was started
        if index_or_alias is None:
            raise BrowserNotFoundError(
                "No valid browser found from: {}".format(
                    ", ".join(browser for browser in browser_options)
                )
            )

        self.go_to(url)
        return index_or_alias

    def _get_browser_order(self, browser_selection: Any) -> list:
        """Get a list of browsers that will be used for open browser
        keywords. Will be one or many.

        ``browser_selection`` ``AUTOMATIC_BROWSER_SELECTION`` will be OS-specific list,
            or one named browser, eg. "Chrome"
        """
        if browser_selection == self.AUTOMATIC_BROWSER_SELECTION:
            preferable_browser_order = webdriver.DRIVER_PREFERENCE.get(
                platform.system(), webdriver.DRIVER_PREFERENCE["default"]
            )
        else:
            preferable_browser_order = (
                browser_selection
                if isinstance(browser_selection, list)
                else browser_selection.split(",")
            )
        return preferable_browser_order

    def _get_driver_args(
        self,
        browser: str,
        headless: bool = False,
        maximized: bool = False,
        use_profile: bool = False,
        profile_name: Optional[str] = None,
        profile_path: Optional[str] = None,
        preferences: Optional[dict] = None,
        proxy: str = None,
    ) -> dict:
        """Get browser and webdriver arguments for given options."""
        preferences = preferences or {}
        browser = browser.lower()
        headless = headless or bool(int(os.getenv("RPA_HEADLESS_MODE", "0")))
        kwargs = {}

        if browser not in self.AVAILABLE_OPTIONS:
            return kwargs, []

        module = importlib.import_module("selenium.webdriver")
        factory = getattr(module, self.AVAILABLE_OPTIONS[browser])
        options = factory()

        if headless:
            self._set_headless_options(browser, options)

        if maximized:
            options.add_argument("--start-maximized")

        if browser != "chrome":
            kwargs["options"] = options
            if use_profile:
                self.logger.warning("Profiles are supported only with Chrome")

        else:
            default_preferences = {
                "safebrowsing.enabled": True,
                "credentials_enable_service": False,
                "profile.password_manager_enabled": False,
            }
            if proxy:
                options.add_argument("--proxy-server=%s" % proxy)
            options.add_argument("--disable-web-security")
            options.add_argument("--allow-running-insecure-content")
            options.add_argument("--no-sandbox")
            options.add_experimental_option(
                "prefs", {**default_preferences, **preferences}
            )
            options.add_experimental_option(
                "excludeSwitches", ["enable-logging", "enable-automation"]
            )

            if use_profile:
                self._set_user_profile(options, profile_path, profile_name)

            if self.logger.isEnabledFor(logging.DEBUG):
                kwargs["service_log_path"] = "chromedriver.log"
                kwargs["service_args"] = ["--verbose"]

            kwargs["chrome_options"] = options

        return kwargs, options.arguments

    def _set_headless_options(self, browser: str, options: dict) -> None:
        """Set headless mode for the browser, if possible.

        ``browser`` string name of the browser

        ``options`` browser options class instance
        """
        if browser.lower() == "safari":
            self.logger.warning(
                "Safari does not support headless mode. "
                "(https://github.com/SeleniumHQ/selenium/issues/5985)"
            )
            return

        options.add_argument("--headless")
        options.add_argument("--disable-gpu")
        options.add_argument("--disable-dev-shm-usage")

    def _set_user_profile(
        self,
        options: dict,
        profile_path: Optional[str] = None,
        profile_name: Optional[str] = None,
    ) -> None:
        """Set user profile configuration into browser options

        Requires environment variable ``RPA_CHROME_USER_PROFILE_DIR``
        to point into user profile directory.

        ``options`` dictionary of browser options
        """
        data_dir = profile_path or os.getenv("RPA_CHROME_USER_PROFILE_DIR")

        system = platform.system()
        home = Path.home()

        if data_dir is not None:
            pass
        elif system == "Windows":
            data_dir = home / "AppData" / "Local" / "Google" / "Chrome" / "User Data"
        elif system == "Linux":
            data_dir = home / ".config" / "google-chrome"
        elif system == "Darwin":
            data_dir = home / "Library" / "Application Support" / "Google" / "Chrome"
        else:
            self.logger.warning("Unable to resolve profile directory for: %s", system)
            return

        if not Path(data_dir).exists():
            self.logger.warning("Given profile directory does not exist: %s", data_dir)

        options.add_argument("--enable-local-sync-backend")
        options.add_argument(f"--local-sync-backend-dir={data_dir}")
        options.add_argument(f"--user-data-dir={data_dir}")

        if profile_name is not None:
            options.add_argument(f"--profile-directory={profile_name}")

    def _create_webdriver(self, browser, alias, download, **kwargs):
        """Create a webdriver instance for the given browser.

        Returns an index/alias of the webdriver session,
        or ``None`` if a webdriver was not initialized.
        """
        executable = webdriver.executable(browser, download)
        if executable:
            kwargs.setdefault("executable_path", executable)

        library = BrowserManagementKeywords(self)
        browser = browser.lower().capitalize()

        return library.create_webdriver(browser, alias, **kwargs)

    @keyword
    def open_chrome_browser(
        self,
        url: str,
        use_profile: bool = False,
        headless: bool = False,
        maximized: bool = False,
        alias: Optional[str] = None,
        profile_name: Optional[str] = None,
        profile_path: Optional[str] = None,
        preferences: Optional[dict] = None,
        proxy: str = None,
    ) -> int:
        """Open Chrome browser. See ``Open Available Browser`` for
        descriptions of arguments.
        """
        return self.open_available_browser(
            url,
            alias=alias,
            headless=headless,
            maximized=maximized,
            use_profile=use_profile,
            browser_selection="Chrome",
            profile_name=profile_name,
            profile_path=profile_path,
            preferences=preferences,
            proxy=proxy,
        )

    @keyword
    def attach_chrome_browser(self, port: int, alias: Optional[str] = None):
        """Attach to an existing instance of Chrome or Chromium.

        Requires that the browser was started with the command line
        option ``--remote-debugging-port=<port>``, where port is any
        4-digit number not being used by other applications.

        That port can then be used to connect using this keyword.

        Example:
            | Attach Chrome Browser | port=9222 |
        """
        options = ChromeOptions()
        options.add_experimental_option("debuggerAddress", f"localhost:{port:d}")
        create = partial(
            self._create_webdriver, "Chrome", alias, chrome_options=options
        )

        try:
            return create(download=False)
        except Exception:  # pylint: disable=broad-except
            self.logger.debug(traceback.format_exc())
        return create(download=True)

    @keyword
    def open_headless_chrome_browser(self, url: str) -> int:
        """Open Chrome browser in headless mode.

        ``url`` URL to open

        Example:
            | ${idx} | Open Headless Chrome Browser | https://www.google.com |
        """
        return self.open_chrome_browser(url, headless=True)

    @keyword
    def screenshot(
        self,
        locator: str = None,
        filename: str = "",
    ) -> None:
        # pylint: disable=C0301, W0212
        """Capture page and/or element screenshot.

        ``locator`` if defined, take element screenshot, if not takes page screenshot

        ``filename`` filename for the screenshot, by default creates file `screenshot-timestamp-element/page.png`
        if set to `None` then file is not saved at all

        Example:
            | Screenshot | locator=//img[@alt="Google"] | filename=locator.png |             # element screenshot, defined filename
            | Screenshot | filename=page.png            |                                    # page screenshot, defined filename
            | Screenshot | filename=${NONE}             |                                    # page screenshot, NO file will be created
            | Screenshot |                              |                                    # page screenshot, default filename
            | Screenshot | locator=//img[@alt="Google"] |                                    # element screenshot, default filename
            | Screenshot | locator=//img[@alt="Google"] | filename=${CURDIR}/subdir/loc.png  # element screenshot, create dirs if not existing
        """  # noqa: E501
        screenshot_keywords = ScreenshotKeywords(self)
        default_filename_prefix = f"screenshot-{int(time.time())}"

        def __save_base64_screenshot_to_file(base64_string, filename):
            path = screenshot_keywords._get_screenshot_path(filename)
            screenshot_keywords._create_directory(path)
            with open(filename, "wb") as fh:
                fh.write(base64.b64decode(base64_string))
                self.logger.info("Screenshot saved to file: %s", filename)

        if locator:
            element = screenshot_keywords.find_element(locator)
            screenshot_keywords._embed_to_log_as_base64(
                element.screenshot_as_base64, 400
            )
            if filename is not None:
                filename = filename or os.path.join(
                    os.curdir, f"{default_filename_prefix}-element.png"
                )
                __save_base64_screenshot_to_file(element.screenshot_as_base64, filename)
        else:
            screenshot_as_base64 = self.driver.get_screenshot_as_base64()
            screenshot_keywords._embed_to_log_as_base64(screenshot_as_base64, 800)
            if filename is not None:
                filename = filename or os.path.join(
                    os.curdir, f"{default_filename_prefix}-page.png"
                )
                __save_base64_screenshot_to_file(screenshot_as_base64, filename)

    @keyword
    def click_element_when_visible(
        self, locator: str, modifier: Optional[str] = None, action_chain: bool = False
    ) -> None:
        """Click element identified by ``locator``, once it becomes visible.

        ``locator`` element locator

        ``modifier`` press given keys while clicking the element, e.g. CTRL

        ``action_chain`` store action in Selenium ActionChain queue

        Example:
            | Click Element When Visible | q |
            | Click Element When Visible | id:button | CTRL+ALT |
            | Click Element When Visible | action_chain=True |
        """
        self.wait_until_element_is_visible(locator)
        self.click_element(locator, modifier, action_chain)

    @keyword
    def click_button_when_visible(
        self, locator: str, modifier: Optional[str] = None
    ) -> None:
        """Click button identified by ``locator``, once it becomes visible.

        ``locator`` element locator

        ``modifier`` press given keys while clicking the element, e.g. CTRL

        Example:
            | Click Button When Visible  | //button[@class="mybutton"] |
        """
        self.wait_until_element_is_visible(locator)
        self.click_button(locator, modifier)

    # Alias for backwards compatibility
    wait_and_click_button = click_button_when_visible

    @keyword
    def click_element_if_visible(self, locator: str) -> None:
        """Click element if it is visible

        ``locator`` element locator

        Example:
            | Click Element If Visible | //button[@class="mybutton"] |
        """
        visible = self.is_element_visible(locator)
        if visible:
            self.click_element(locator)

    @keyword
    def input_text_when_element_is_visible(self, locator: str, text: str) -> None:
        # pylint: disable=C0301
        """Input text into locator after it has become visible.

        ``locator`` element locator

        ``text`` insert text to locator

        Example:
            | Input Text When Element Is Visible | //input[@id="freetext"]  | my feedback |
        """  # noqa: E501
        self.wait_until_element_is_visible(locator)
        self.input_text(locator, text)

    @keyword
    def is_element_enabled(self, locator: str) -> bool:
        """Is element enabled

        ``locator`` element locator

        Example:
            | ${res} | Is Element Enabled | input.field1 |
        """
        return self._run_should_keyword_and_return_status(
            self.element_should_be_enabled, locator
        )

    @keyword
    def is_element_visible(self, locator: str) -> bool:
        """Is element visible

        ``locator`` element locator

        Example:
            | ${res} | Is Element Visible | id:confirmation |
        """
        return self._run_should_keyword_and_return_status(
            self.element_should_be_visible, locator
        )

    @keyword
    def is_element_disabled(self, locator: str) -> bool:
        """Is element disabled

        ``locator`` element locator

        Example:
            | ${res} | Is Element Disabled | //input[@type="submit"] |
        """
        return self._run_should_keyword_and_return_status(
            self.element_should_be_disabled, locator
        )

    @keyword
    def is_element_focused(self, locator: str) -> bool:
        """Is element focused

        ``locator`` element locator

        Example:
            | ${res} | Is Element Focused | //input[@id="freetext"] |
        """
        return self._run_should_keyword_and_return_status(
            self.element_should_be_focused, locator
        )

    @keyword
    def is_element_attribute_equal_to(
        self, locator: str, attribute: str, expected: str
    ) -> bool:
        """Is element attribute equal to expected value

        ``locator`` element locator

        ``attribute`` element attribute to check for

        ``expected`` is attribute value equal to this

        Example:
            | ${res} | Is Element Attribute Equal To | h1 | id | main |
        """
        return self._run_should_keyword_and_return_status(
            self.element_attribute_value_should_be, locator, attribute, expected
        )

    @keyword
    def is_alert_present(self, text: str = None, action: str = "ACCEPT") -> bool:
        """Is alert box present, which can be identified with text
        and action can also be done which by default is ACCEPT.

        Other possible actions are DISMISS and LEAVE.

        ``text`` check if alert text is matching to this, if `None`
        will check if alert is present at all

        ``action`` possible action if alert is present, default ACCEPT

        Example:
            | ${res} | Is Alert Present | id:cookies |
        """
        return self._run_should_keyword_and_return_status(
            self.alert_should_be_present, text, action
        )

    @keyword
    def is_checkbox_selected(self, locator: str) -> bool:
        """Is checkbox selected

        ``locator`` element locator

        Example:
            | ${res} |  Is Checkbox Selected  | id:taxes-paid |
        """
        return self._run_should_keyword_and_return_status(
            self.checkbox_should_be_selected, locator
        )

    @keyword
    def does_frame_contain(self, locator: str, text: str) -> bool:
        """Does frame contain expected text

        ``locator`` locator of the frame to check

        ``text`` does frame contain this text

        Example:
            | ${res} | Does Frame Contain | id:myframe | secret |
        """
        return self._run_should_keyword_and_return_status(
            self.frame_should_contain, locator, text
        )

    @keyword
    def does_element_contain(
        self, locator: str, expected: str, ignore_case: bool = False
    ) -> bool:
        # pylint: disable=C0301
        """Does element contain expected text

        ``locator`` element locator

        ``expected`` expected element text

        ``ignore_case`` should check be case insensitive, default `False`

        Example:
            | ${res} | Does Element Contain | id:spec | specification complete | ignore_case=True |
        """  # noqa: E501
        return self._run_should_keyword_and_return_status(
            self.element_should_contain,
            locator=locator,
            expected=expected,
            ignore_case=ignore_case,
        )

    @keyword
    def is_element_text(
        self, locator: str, expected: str, ignore_case: bool = False
    ) -> bool:
        """Is element text expected

        ``locator`` element locator

        ``expected`` expected element text

        ``ignore_case`` should check be case insensitive, default `False`

        Example:
            | ${res} | Is Element Text | id:name | john doe |
            | ${res} | Is Element Text | id:name | john doe | ignore_case=True |
        """
        return self._run_should_keyword_and_return_status(
            self.element_text_should_be,
            locator=locator,
            expected=expected,
            ignore_case=ignore_case,
        )

    @keyword
    def is_list_selection(self, locator: str, *expected: str) -> bool:
        """Is list selected with expected values

        ``locator`` element locator

        ``expected`` expected selected options

        Example:
            | ${res} | Is List Selection | id:cars | Ford |
        """
        return self._run_should_keyword_and_return_status(
            self.list_selection_should_be, locator, *expected
        )

    @keyword
    def is_list_selected(self, locator: str) -> bool:
        """Is any option selected in the

        ``locator`` element locator

        Example:
            | ${res} | Is List Selected | id:cars |
        """
        self.logger.info("Will return if anything is selected on the list")
        return not self._run_should_keyword_and_return_status(
            self.list_should_have_no_selections, locator
        )

    @keyword
    def is_location(self, url: str) -> bool:
        """Is current URL expected url

        ``url`` expected current URL

        Example:
            | Open Available Browser | https://www.robocorp.com |
            | ${res} | Is Location | https://www.robocorp.com |
        """
        return self._run_should_keyword_and_return_status(self.location_should_be, url)

    @keyword
    def does_location_contain(self, expected: str) -> bool:
        """Does current URL contain expected

        ``expected`` URL should contain this

        Example:
            | Open Available Browser | https://robocorp.com |
            | ${res} | Does Location Contain | robocorp |
        """
        return self._run_should_keyword_and_return_status(
            self.location_should_contain, expected
        )

    @keyword
    def does_page_contain(self, text: str) -> bool:
        """Does page contain expected text

        ``text`` page should contain this

        Example:
            | Open Available Browser | https://google.com |
            | ${res} | Does Page Contain | Gmail |
        """
        return self._run_should_keyword_and_return_status(
            self.page_should_contain, text
        )

    @keyword
    def does_page_contain_button(self, locator: str) -> bool:
        """Does page contain expected button

        ``locator`` element locator

        Example:
            | ${res} | Does Page Contain Button | search-button |
        """
        return self._run_should_keyword_and_return_status(
            self.page_should_contain_button, locator
        )

    @keyword
    def does_page_contain_checkbox(self, locator: str) -> bool:
        """Does page contain expected checkbox

        ``locator`` element locator

        Example:
            | ${res} | Does Page Contain Checkbox | random-selection |
        """
        return self._run_should_keyword_and_return_status(
            self.page_should_contain_checkbox, locator
        )

    @keyword
    def does_page_contain_element(self, locator: str, count: int = None) -> bool:
        """Does page contain expected element

        ``locator`` element locator

        ``count`` how many times element is expected to appear on page
        by default one or more

        Example:
            | ${res} | Does Page Contain Element | textarea |
            | ${res} | Does Page Contain Element | button | count=4 |
        """
        return self._run_should_keyword_and_return_status(
            self.page_should_contain_element, locator=locator, limit=count
        )

    @keyword
    def does_page_contain_image(self, locator: str) -> bool:
        """Does page contain expected image

        ``locator`` element locator

        Example:
            | Open Available Browser | https://google.com |
            | ${res} | Does Page Contain Image | Google |
        """
        return self._run_should_keyword_and_return_status(
            self.page_should_contain_image, locator
        )

    @keyword
    def does_page_contain_link(self, locator: str) -> bool:
        """Does page contain expected link

        ``locator`` element locator

        Example:
            | ${res} | Does Page Contain Link | id:submit |
        """
        return self._run_should_keyword_and_return_status(
            self.page_should_contain_link, locator
        )

    @keyword
    def does_page_contain_list(self, locator: str) -> bool:
        """Does page contain expected list

        ``locator`` element locator

        Example:
            | ${res} | Does Page Contain List | class:selections |
        """
        return self._run_should_keyword_and_return_status(
            self.page_should_contain_list, locator
        )

    @keyword
    def does_page_contain_radio_button(self, locator: str) -> bool:
        """Does page contain expected radio button

        ``locator`` element locator

        Example:
            | ${res} | Does Page Contain Radio Button | male |
        """
        return self._run_should_keyword_and_return_status(
            self.page_should_contain_radio_button, locator
        )

    @keyword
    def does_page_contain_textfield(self, locator: str) -> bool:
        """Does page contain expected textfield

        ``locator`` element locator

        Example:
            | ${res} | Does Page Contain Textfield | id:address |
        """
        return self._run_should_keyword_and_return_status(
            self.page_should_contain_textfield, locator
        )

    @keyword
    def is_radio_button_set_to(self, group_name: str, value: str) -> bool:
        """Is radio button group set to expected value

        ``group_name`` radio button group name

        ``value`` expected value

        Example:
            | ${res} | Is Radio Button Set To | group_name=gender | value=female |
        """
        return self._run_should_keyword_and_return_status(
            self.radio_button_should_be_set_to, group_name, value
        )

    @keyword
    def is_radio_button_selected(self, group_name: str) -> bool:
        """Is any radio button selected in the button group

        ``group_name`` radio button group name

        Example:
            | ${res} | Is Radio Button Selected | group_name=gender |
        """
        self.logger.info(
            "Will return if anything is selected on the radio button group"
        )
        return not self._run_should_keyword_and_return_status(
            self.radio_button_should_not_be_selected, group_name
        )

    @keyword
    def does_table_cell_contain(
        self, locator: str, row: int, column: int, expected: str
    ) -> bool:
        """Does table cell contain expected text

        ``locator`` element locator for the table

        ``row`` row index starting from 1 (beginning) or -1 (from the end)

        ``column`` column index starting from 1 (beginning) or -1 (from the end)

        ``expected`` expected text in table row

        Example:
            | ${res} | Does Table Cell Contain | //table | 1 | 1 | Company |
        """
        return self._run_should_keyword_and_return_status(
            self.table_cell_should_contain, locator, row, column, expected
        )

    @keyword
    def does_table_column_contain(
        self, locator: str, column: int, expected: str
    ) -> bool:
        """Does table column contain expected text

        ``locator`` element locator for the table

        ``column`` column index starting from 1 (beginning) or -1 (from the end)

        ``expected`` expected text in table column

        Example:
            | ${res} | Does Table Column Contain | //table | 1 | Nokia |
        """
        return self._run_should_keyword_and_return_status(
            self.table_column_should_contain, locator, column, expected
        )

    @keyword
    def does_table_row_contain(self, locator: str, row: int, expected: str) -> bool:
        """Does table row contain expected text

        ``locator`` element locator for the table

        ``row`` row index starting from 1 (beginning) or -1 (from the end)

        ``expected`` expected text in table row

        Example:
            | ${res} | Does Table Row Contain | //table | 1 | Company |
        """
        return self._run_should_keyword_and_return_status(
            self.table_row_should_contain, locator, row, expected
        )

    @keyword
    def does_table_footer_contain(self, locator: str, expected: str) -> bool:
        """Does table footer contain expected text

        ``locator`` element locator for the table

        ``expected`` expected text in table footer

        Example:
            | ${res} | Does Table Footer Contain | //table | Sum |
        """
        return self._run_should_keyword_and_return_status(
            self.table_footer_should_contain, locator, expected
        )

    @keyword
    def does_table_header_contain(self, locator: str, expected: str) -> bool:
        """Does table header contain expected text

        ``locator`` element locator for the table

        ``expected`` expected text in table header

        Example:
            | ${res} | Does Table Header Contain | //table | Month |
        """
        return self._run_should_keyword_and_return_status(
            self.table_header_should_contain, locator, expected
        )

    @keyword
    def does_table_contain(self, locator: str, expected: str) -> bool:
        """Does table contain expected text

        ``locator`` element locator

        ``expected`` expected text in table

        Example:
            | ${res} | Does Table Contain | //table | February |
        """
        return self._run_should_keyword_and_return_status(
            self.table_should_contain, locator, expected
        )

    @keyword
    def is_textarea_value(self, locator: str, expected: str) -> bool:
        """Is textarea matching expected value

        ``locator`` element locator

        ``expected`` expected textarea value

        Example:
            | ${res} | Is Textarea Value | //textarea | Yours sincerely |
        """
        return self._run_should_keyword_and_return_status(
            self.textarea_value_should_be, locator, expected
        )

    @keyword
    def does_textarea_contain(self, locator: str, expected: str) -> bool:
        """Does textarea contain expected text

        ``locator`` element locator

        ``expected`` expected text in textarea

        Example:
            | ${res} | Does Textarea Contain | //textarea | sincerely |
        """
        return self._run_should_keyword_and_return_status(
            self.textarea_should_contain, locator, expected
        )

    @keyword
    def does_textfield_contain(self, locator: str, expected: str) -> bool:
        """Does textfield contain expected text

        ``locator`` element locator

        ``expected`` expected text in textfield

        Example:
            | ${res} | Does Textfield Contain | id:lname | Last |
        """
        return self._run_should_keyword_and_return_status(
            self.textfield_should_contain, locator, expected
        )

    @keyword
    def is_textfield_value(self, locator: str, expected: str) -> bool:
        """Is textfield value expected

        ``locator`` element locator

        ``expected`` expected textfield value

        Example:
            | ${res} | Is Textfield Value | id:lname | Lastname |
        """
        return self._run_should_keyword_and_return_status(
            self.textfield_value_should_be, locator, expected
        )

    @keyword
    def is_title(self, title: str) -> bool:
        """Is page title expected

        ``title`` expected title value

        Example:
            | ${res} | Is Title | Webpage title text |
        """
        return self._run_should_keyword_and_return_status(self.title_should_be, title)

    def _run_should_keyword_and_return_status(self, runnable_keyword, *args, **kwargs):
        try:
            runnable_keyword(*args, **kwargs)
            return True
        except AssertionError:
            return False

    @keyword
    def get_element_status(self, locator: str) -> dict:
        """Return dictionary containing element status of:

            - visible
            - enabled
            - disabled
            - focused

        ``locator`` element locator

        Example:
            | &{res}  | Get Element Status | class:special |
            | Log     | ${res.visible} |
            | Log     | ${res.enabled} |
            | Log     | ${res.disabled} |
            | Log     | ${res.focused} |
        """
        status_object = dict()
        status_object["visible"] = self.is_element_visible(locator)
        status_object["enabled"] = self.is_element_enabled(locator)
        status_object["disabled"] = self.is_element_disabled(locator)
        status_object["focused"] = self.is_element_focused(locator)
        notebook.notebook_json(status_object)
        return status_object

    @keyword
    def get_testability_status(self) -> bool:
        """Get SeleniumTestability plugin status"""
        return self.using_testability



import logging
from DatabaseLibrary import DatabaseLibrary

from RPA.Tables import Table


class Database(DatabaseLibrary):
    """Library handling different database operations.

    Extends functionality of DatabaseLibrary, for more information see
    https://github.com/franz-see/Robotframework-Database-Library
    """

    def __init__(self, *args, **kwargs) -> None:
        self.logger = logging.getLogger(__name__)
        DatabaseLibrary.__init__(self, *args, **kwargs)

    def database_query_result_as_table(self, query_string: str) -> Table:
        """Make database query and return result as `RPA.Table`.

        ``query_string`` SQL query string
        """
        return Table(self.query(query_string, returnAsDict=True))



"""
WorkItem library skeleton

* Multiple adapters for different backend
* Default endpoints, tokens, ids from env (runtime)
"""
import json
import logging
from abc import ABC, abstractmethod

import requests
from requests.exceptions import HTTPError
from robot.libraries.BuiltIn import BuiltIn

from RPA.core.helpers import import_by_name, required_env
from RPA.core.notebook import notebook_print


def json_dump_safe(data, **kwargs):
    """Convert data to JSON string, and skip invalid values."""

    def invalid(obj):
        name = type(obj).__qualname__
        logging.warning("Failed to serialize: %s", name)
        return f"<non-serializable: {name}>"

    if data is None:
        data = {}

    return json.dumps(data, default=invalid, **kwargs)


class BaseAdapter(ABC):
    """Abstract base class for work item adapters."""

    @abstractmethod
    def load(self, workspace_id, item_id):
        raise NotImplementedError

    @abstractmethod
    def save(self, workspace_id, item_id, data):
        raise NotImplementedError


class RobocloudAdapter(BaseAdapter):
    """Adapter for saving/loading work items from Robocloud.

    Required environment variables:

    * RC_API_WORKITEM_HOST:     Work item data API hostname
    * RC_API_WORKITEM_TOKEN:    Work item data API access token
    """

    def __init__(self):
        self.host = required_env("RC_API_WORKITEM_HOST")
        self.token = required_env("RC_API_WORKITEM_TOKEN")

    @property
    def headers(self):
        return {"Authorization": f"Bearer {self.token}"}

    def url(self, workspace_id, item_id):
        return f"{self.host}/json-v1/workspaces/{workspace_id}/workitems/{item_id}/data"

    def handle_response(self, request):
        if request.ok:
            return request.json()

        if request.status_code == 404:
            return {}

        try:
            response = request.json()
        except ValueError:
            request.raise_for_status()

        status_code = response.get("status", request.status_code)
        status_msg = response.get("error", {}).get("code", "Error")
        reason = response.get("message") or response.get("error", {}).get(
            "message", request.reason
        )

        raise HTTPError(f"{status_code} {status_msg}: {reason}")

    def load(self, workspace_id, item_id):
        url = self.url(workspace_id, item_id)
        logging.info("Loading item from %s", url)

        resp = requests.get(url, headers=self.headers)
        return self.handle_response(resp)

    def save(self, workspace_id, item_id, data):
        url = self.url(workspace_id, item_id)
        logging.info("Saving item to %s", url)

        data = json_dump_safe(data)
        logging.info("Payload: %s", data)

        resp = requests.put(url, headers=self.headers, data=data)
        return self.handle_response(resp)


class FileAdapter(BaseAdapter):
    """Adapter for saving/loading work items from disk.

    Required environment variables:

    * RPA_WORKITEMS_PATH:   Path to work items database file
    """

    def __init__(self):
        self.path = required_env("RPA_WORKITEMS_PATH")

    def _read(self):
        try:
            with open(self.path, "r") as infile:
                return json.load(infile)
        except IOError as err:
            logging.info("Failed to read database: %s", err)
            return {}

    def load(self, workspace_id, item_id):
        content = self._read()
        try:
            return content[workspace_id][item_id]
        except KeyError:
            return {}

    def save(self, workspace_id, item_id, data):
        content = self._read()
        content.setdefault(workspace_id, {})[item_id] = data

        with open(self.path, "w") as outfile:
            outfile.write(json_dump_safe(content, indent=4))


class WorkItem:
    """Container for a single work item.

    :param workspace_id:    Workspace ID which contains item
    :param item_id:         Workitem ID
    :param adapter:         Adapter for storage backend
    """

    def __init__(self, workspace_id, item_id, adapter):
        self.workspace_id = workspace_id
        self.item_id = item_id
        self.adapter = adapter

        #: Current item payload
        self.data = None

    def __str__(self):
        return f"WorkItem(workspace={self.workspace_id},id={self.item_id})"

    def __enter__(self):
        self.load()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            self.save()

    def show(self):
        return "{item}: {data}".format(item=self, data=json.dumps(self.data, indent=4))

    def load(self):
        self.data = self.adapter.load(self.workspace_id, self.item_id)
        return self.data

    def save(self):
        self.adapter.save(self.workspace_id, self.item_id, self.data)


class Items:
    """A library for interacting with RPA work items.

    `Items` is a collection of keywords for handling data
    that is moved between different processes and Robot Framework
    tasks. It allows storing and restoring values to/from cloud or file based
    storage, and manipulating their contents.

    :param load_env: Automatically load work item using environment variables
    :param default_adapter: Set default adapter if not overriden by environment

    Environment variables:

    * RPA_WORKITEMS_ADAPTER: Import path to adapter, e.g. "mymodule.MyAdapter"
    * RC_WORKSPACE_ID:       Current default workspace ID
    * RC_WORKITEM_ID:        Current default work item ID
    """

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LISTENER_API_VERSION = 2

    def __init__(self, load_env=True, default_adapter=RobocloudAdapter):
        self.ROBOT_LIBRARY_LISTENER = self
        #: Adapter factory for new work items
        self.adapter = None
        #: The current active work item, set automatically to latest loaded item
        self.current = None
        self._load_env = load_env
        self._load_adapter(default_adapter)

    def _load_adapter(self, default):
        """Load adapter by name, using env or given default."""
        adapter = required_env("RPA_WORKITEMS_ADAPTER", default)
        if isinstance(adapter, str):
            self.adapter = import_by_name(adapter, __name__)
        else:
            self.adapter = adapter
        assert issubclass(
            self.adapter, BaseAdapter
        ), "Adapter does not inherit from BaseAdapter"

    def _start_suite(self, data, result):
        """Robot Framework listener method, called when suite starts."""
        # pylint: disable=unused-argument, broad-except
        if self._load_env:
            try:
                self.load_work_item_from_environment()
            except Exception as exc:
                logging.warning("Failed to load item: %s", exc)
            finally:
                self._load_env = False

    def load_work_item_from_environment(self):
        """Load current work item defined by the runtime environment.

        The corresponding environment variables are:

        * RC_WORKSPACE_ID
        * RC_WORKITEM_ID
        """
        workspace_id = required_env("RC_WORKSPACE_ID")
        item_id = required_env("RC_WORKITEM_ID")
        return self.load_work_item(workspace_id, item_id)

    def load_work_item(self, workspace_id, item_id):
        """Load work item for reading/writing.

        :param workspace_id:    Workspace ID which contains item
        :param item_id:         Workitem ID to load
        """
        item = WorkItem(workspace_id, item_id, self.adapter())
        item.load()
        self.current = item
        return self.current

    def save_work_item(self):
        """Save the current data in the work item. If not saved,
        all changes are discarded when the library goes out of scope.
        """
        assert self.current, "No active work item"
        self.current.save()

    def clear_work_item(self):
        """Remove all data in the current work item."""
        assert self.current, "No active work item"
        self.current.data = {}

    def get_work_item_payload(self):
        """Get the full JSON payload for a work item.

        NOTE: Most use cases should prefer higher-level keywords.
        """
        assert self.current, "No active work item"
        return self.current.data

    def set_work_item_payload(self, payload):
        """Set the full JSON payload for a work item.

        NOTE: Most use cases should prefer higher-level keywords.

        :param payload: Content of payload, must be JSON-serializable
        """
        assert self.current, "No active work item"
        self.current.data = payload

    def list_work_item_variables(self):
        """List the variable names for the current work item."""
        return list(self.get_work_item_variables().keys())

    def get_work_item_variables(self):
        """Read all variables from the current work item and
        return their names and values as a dictionary.
        """
        assert self.current, "No active work item"
        return self.current.data.setdefault("variables", {})

    def get_work_item_variable(self, name, default=None):
        """Return a single variable value from the work item,
        or default value if defined and key does not exist.
        If key does not exist and default is not defined, raises `KeyError`.

        :param key:     Name of variable
        :param default: Default value if key does not exist
        """
        variables = self.get_work_item_variables()
        value = variables.get(name, default)
        if value is None:
            raise KeyError(f"Undefined variable: {name}")
        notebook_print(text=f"**{name}** = **{value}**")
        return value

    def set_work_item_variables(self, **kwargs):
        """Set multiple variables in the current work item.

        :param kwargs: Pairs of variable names and values
        """
        variables = self.get_work_item_variables()
        for name, value in kwargs.items():
            logging.info("%s = %s", name, value)
            variables[name] = value

    def set_work_item_variable(self, name, value):
        """Set a single variable value in the current work item.

        :param key:     Name of variable
        :param value:   Value of variable
        """
        variables = self.get_work_item_variables()
        logging.info("%s = %s", name, value)
        variables[name] = value

    def delete_work_item_variables(self, *names, force=True):
        """Delete variable(s) from the current work item.

        :param names:  names of variables to remove
        :param force:  ignore variables that don't exist in work item
        """
        variables = self.get_work_item_variables()
        for name in names:
            if name in variables:
                del variables[name]
                logging.info("Deleted variable: %s", name)
            elif not force:
                raise KeyError(f"No such variable: {name}")

    def set_task_variables_from_work_item(self):
        """Convert all variables in the current work item to
        Robot Framework task variables.
        """
        variables = self.get_work_item_variables()
        for name, value in variables.items():
            BuiltIn().set_task_variable(f"${{{name}}}", value)



# pylint: disable=c-extension-no-member
import logging
import platform

import clipboard

if platform.system() == "Windows":
    import win32clipboard


class Clipboard:
    """RPA Framework library for cross platform clipboard management.

    Will use `win32` package on Windows and `clipboard` package on Linux and Mac.
    """

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    def copy_to_clipboard(self, text):
        """Copy text to clipboard

        :param text: to copy
        """
        self.logger.debug("copy_to_clipboard")
        if platform.system() == "Windows":
            win32clipboard.OpenClipboard()
            win32clipboard.EmptyClipboard()
            win32clipboard.SetClipboardText(text)
            win32clipboard.CloseClipboard()
        else:
            clipboard.copy(text)

    def paste_from_clipboard(self):
        """Paste text from clipboard

        :return: text
        """
        self.logger.debug("paste_from_clipboard")
        if platform.system() == "Windows":
            win32clipboard.OpenClipboard()
            if win32clipboard.IsClipboardFormatAvailable(win32clipboard.CF_TEXT):
                text = win32clipboard.GetClipboardData()
            else:
                text = None
            win32clipboard.CloseClipboard()
            return text
        else:
            return clipboard.paste()

    def clear_clipboard(self):
        """Clear clipboard contents"""
        self.logger.debug("clear_clipboard")
        if platform.system() == "Windows":
            win32clipboard.OpenClipboard()
            win32clipboard.EmptyClipboard()
            win32clipboard.CloseClipboard()
        else:
            clipboard.copy("")
